@@ -1,5 +1,5 @@
 /*  GNU Ocrad - Optical Character Recognition program
-    Copyright (C) 2003, 2004 Antonio Diaz Diaz.
+    Copyright (C) 2003, 2004, 2005 Antonio Diaz Diaz.
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -13,7 +13,7 @@
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
-    Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
 
 #include <cstdio>
@@ -21,8 +21,8 @@
 #include "common.h"
 #include "rectangle.h"
 #include "ucs.h"
+#include "bitmap.h"
 #include "block.h"
-#include "blockmap.h"
 #include "character.h"
 #include "profile.h"
 #include "feats.h"
@@ -30,7 +30,8 @@
 
 // Recognizes 2 block characters.
 // ij!%:;=?|
-void Character::recognize12( const Charset & charset, int charbox_vcenter ) throw()
+//
+void Character::recognize12( const Charset & charset, const Rectangle & charbox ) throw()
   {
   const Block & b1 = block( 0 );		// upper block
   const Block & b2 = block( 1 );		// lower block
@@ -41,7 +42,7 @@ void Character::recognize12( const Charset & charset, int charbox_vcenter ) thro
 
   if( Ocrad::similar( a1, a2, 10 ) )
     {
-    if( !b1.blocks() && !b2.blocks() &&
+    if( !b1.holes() && !b2.holes() &&
         2 * a1 > b1.size() && 2 * a2 > b2.size() )
       {
       if( width() > height() || Ocrad::similar( width(), height(), 40 ) )
@@ -55,22 +56,22 @@ void Character::recognize12( const Charset & charset, int charbox_vcenter ) thro
     }
   if( Ocrad::similar( a1, a2, 60 ) )
     {
-    if( f1.test_solid( charbox_vcenter ) == '.' )
+    if( f1.test_solid( charbox ) == '.' )
       {
-      if( f2.test_solid( charbox_vcenter ) == '.' )
+      if( f2.test_solid( charbox ) == '.' )
         { add_guess( ':', 0 ); return; }
-      if( b2.height() > b1.height() && b2.top() > charbox_vcenter )
+      if( b2.height() > b1.height() && b2.top() > charbox.vcenter() )
         { add_guess( ';', 0 ); return; }
       }
     if( charset.enabled( Charset::iso_8859_15 ) ||
         charset.enabled( Charset::iso_8859_9 ) )
       {
-      int code = f2.test_solid( charbox_vcenter );
+      int code = f2.test_solid( charbox );
       if( code == '-' || code == '_' )
         { add_guess( UCS::PLUSMIN, 0 ); return; }
       }
     if( b1.includes_hcenter( b2 ) && b2.includes_hcenter( b1 ) )
-      { if( b1.blocks() && b2.blocks() ) { add_guess( 'g', 0 ); return; } }
+      { if( b1.holes() && b2.holes() ) { add_guess( 'g', 0 ); return; } }
     if( b1.hcenter() < b2.hcenter() )
       {
       // Looks for merged 'fi'
@@ -78,7 +79,7 @@ void Character::recognize12( const Charset & charset, int charbox_vcenter ) thro
           b1.includes_hcenter( b2 ) && 4 * b1.height() > 5 * b2.height() &&
           Ocrad::similar( b1.bottom()-b1.top(), b2.bottom()-b1.top(), 10 ) )
         {
-        Character c2( b2 ); c2.recognize1( charset, vcenter() );
+        Character c2( new Block( b2 ) ); c2.recognize1( charset, charbox );
         if( c2.maybe('l') || c2.maybe('|') )
           {
           only_guess( 0, b1.left() ); add_guess( 'f', b2.left() );
@@ -86,6 +87,7 @@ void Character::recognize12( const Charset & charset, int charbox_vcenter ) thro
           }
         }
       if( 2 * b1.height() > 3 * b2.height() &&
+          b1.holes() == 1 && b2.holes() == 1 &&
           Ocrad::similar( b2.width(), b2.height(), 50 ) )
         { add_guess( '%', 0 ); return; }
       }
@@ -93,24 +95,26 @@ void Character::recognize12( const Charset & charset, int charbox_vcenter ) thro
   if( a1 < a2 )
     {
       {
-      int code = f1.test_solid( charbox_vcenter );	//FIXME all this
+      int code = f1.test_solid( charbox );		//FIXME all this
       if( code == '-' && 2 * b1.height() > b1.width() ) code = '.';
       else if( code == '\'' || code == '|' ) code = '.';
-      if( !code && !b1.blocks() && 
+      if( !code && !b1.holes() && 
           2 * b1.height() < b2.height() && b1.width() <= b2.width() )
         {
         if( 10 * a1 >= 7 * b1.height() * b1.width() ) code = '.';
         else code = '\'';
         }
-      if( !b2.blocks() && ( code == '.' || code == '\'' ) )
+      if( !b2.holes() && ( code == '.' || code == '\'' ) )
         {
         // Looks for merged 'ri' or 'r'
         if( f2.bp.minima( b2.height() / 4 ) == 2 &&
             b2.top() > b1.bottom() && b2.hcenter() < b1.left() )
           {
-          Character c2( b2 ); c2.recognize1( charset, vcenter() );
+          Character c2( new Block( b2 ) ); c2.recognize1( charset, charbox );
           if( c2.maybe('n') )
             {
+            if( code == '.' && ( b1.left() < b2.hcenter() || b1.right() > b2.right() ) )
+              { add_guess( 'n', 0 ); return; }		// FIXME remove dot
             int col, limit = b2.seek_right( b2.vcenter(), b2.hcenter() );
             for( col = b2.hcenter(); col <= limit; ++col )
               if( b2.seek_bottom( b2.vcenter(), col ) < b2.bottom() ) break;
@@ -136,7 +140,7 @@ void Character::recognize12( const Charset & charset, int charbox_vcenter ) thro
           int hdiff;
           if( b2.bottom_hook( &hdiff ) && std::abs( hdiff ) >= b2.height() / 2 )
             {
-            if( hdiff > 0 && f2.rp.increasing( f2.rp.pos( 75 ) ) )
+            if( hdiff > 0 && f2.rp.increasing( f2.rp.pos( 80 ) ) )
               { add_guess( 'j', 0 ); return; }
             if( hdiff < 0 )
               {
@@ -173,9 +177,9 @@ void Character::recognize12( const Charset & charset, int charbox_vcenter ) thro
       }
 
     // upper block has no holes
-    if( ( b1.bottom() < b2.vcenter() || 2 * a1 < a2 ) && !b1.blocks() )
+    if( ( b1.bottom() < b2.vcenter() || 2 * a1 < a2 ) && !b1.holes() )
       {
-      Character c( b2 ); c.recognize1( charset, c.vcenter() );
+      Character c( new Block( b2 ) ); c.recognize1( charset, charbox );
       if( c.guesses() )
         {
         int code = c.guess( 0 ).code;
@@ -183,7 +187,7 @@ void Character::recognize12( const Charset & charset, int charbox_vcenter ) thro
           {
           int atype = '\'';
           if( UCS::isvowel( code ) &&
-              b1.width() > 2 * b1.height() && f1.hp.iscpit() )
+              2 * b1.width() > 3 * b1.height() && f1.hp.iscpit() )
             atype = ':';
           else if( f1.bp.minima() == 2 ) atype = '^';
           else
@@ -205,8 +209,8 @@ void Character::recognize12( const Charset & charset, int charbox_vcenter ) thro
 
   if( b1.bottom() <= b2.top() )
   {
-  int code = f2.test_solid( charbox_vcenter );
-  if( !b1.blocks() && ( code == '.' || 
+  int code = f2.test_solid( charbox );
+  if( !b1.holes() && ( code == '.' || 
         ( code && Ocrad::similar( b2.height(), b2.width(), 50 ) ) ) )
     {
     if( Ocrad::similar( b1.width(), b2.width(), 50 ) )
@@ -217,9 +221,9 @@ void Character::recognize12( const Charset & charset, int charbox_vcenter ) thro
     if( charset.enabled( Charset::iso_8859_15 ) ||
         charset.enabled( Charset::iso_8859_9 ) )
       {
-      if( b1.blocks() == 1 )
+      if( b1.holes() == 1 )
         {
-        const Block & h = b1.block( 0 );
+        const Bitmap & h = b1.hole( 0 );
         if( Ocrad::similar( h.left() - b1.left(), b1.right() - h.right(), 40 ) )
           { add_guess( UCS::MASCORD, 0 ); return; }
         add_guess( UCS::FEMIORD, 0 ); return;
