@@ -1,5 +1,5 @@
 /*  GNU Ocrad - Optical Character Recognition program
-    Copyright (C) 2003, 2004 Antonio Diaz Diaz.
+    Copyright (C) 2003, 2004, 2005 Antonio Diaz Diaz.
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -13,7 +13,7 @@
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
-    Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
 
 #include <cstdio>
@@ -21,8 +21,8 @@
 #include "common.h"
 #include "rectangle.h"
 #include "ucs.h"
+#include "bitmap.h"
 #include "block.h"
-#include "blockmap.h"
 #include "character.h"
 #include "profile.h"
 #include "feats.h"
@@ -30,14 +30,16 @@
 
 // Recognizes 3 block characters.
 // %
-void Character::recognize13( const Charset & charset ) throw()
+//
+void Character::recognize13( const Charset & charset, const Rectangle & charbox ) throw()
   {
   const Block & b1 = block( 0 );
   const Block & b2 = block( 1 );
   const Block & b3 = block( 2 );		// lower block
-  Character c( b3 );
-  c.recognize1( charset, c.vcenter() );
+  Character c( new Block( b3 ) );
   int code = 0;
+
+  c.recognize1( charset, charbox );
   if( c.guesses() )
     {
     if( c.maybe('.') )
@@ -50,6 +52,12 @@ void Character::recognize13( const Charset & charset ) throw()
              Ocrad::similar( b1.height(), b2.height(), 20 ) &&
              2 * b1.height() < b3.height() )
       code = UCS::compose( c.guess( 0 ).code, ':' );
+    else if( c.maybe('o') )
+      {
+      if( ( b1.hcenter() < b2.hcenter() && b1.holes() == 1 && !b2.holes() ) ||
+          ( b2.hcenter() < b1.hcenter() && b2.holes() == 1 && !b1.holes() ) )
+        code = '%';
+      }
     }
   if( charset.only( Charset::ascii ) )
     {
