@@ -1,5 +1,5 @@
 /*  GNU Ocrad - Optical Character Recognition program
-    Copyright (C) 2003, 2004 Antonio Diaz Diaz.
+    Copyright (C) 2003, 2004, 2005 Antonio Diaz Diaz.
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -13,7 +13,7 @@
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
-    Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
 
 namespace Ocrad
@@ -34,7 +34,7 @@ public:
   bool enable( const char * name ) throw();
   bool enabled( Value cset ) const throw();
   bool only( Value cset ) const throw();
-  void show_error( const char * arg ) const throw();
+  void show_error( const char * program_name, const char * arg ) const throw();
   };
 
 
@@ -50,7 +50,7 @@ public:
   Transformation() throw() : _type( none ) {}
   bool set( const char * name ) throw();
   Type type() const throw() { return _type; }
-  void show_error( const char * arg ) const throw();
+  void show_error( const char * program_name, const char * arg ) const throw();
   };
 
 
