@@ -13,7 +13,7 @@
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
-    Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
 
 #include <algorithm>
@@ -22,24 +22,24 @@
 #include "common.h"
 #include "rectangle.h"
 #include "ucs.h"
+#include "bitmap.h"
 #include "block.h"
-#include "blockmap.h"
 #include "profile.h"
 #include "feats.h"
 
 
 // Looks for three black sections in column hcenter()  1, then tests if
 // upper and lower gaps are open to the right or to the left
+//
 int Features::test_235Esz( const Charset & charset ) const throw()
   {
   static const int csize = 9;
-  static const int ucoff[csize] = { 0, +1,  0, -1,  0, +1, -1, +1, -1 };
-  static const int lcoff[csize] = { 0,  0, -1,  0, +1, -1, +1, +1, -1 };
+  static const int ucoff[csize] = { 0, +1,  0, -1,  0, +1, +1, -1, -1 };
+  static const int lcoff[csize] = { 0,  0, -1,  0, +1, -1, +1, -1, +1 };
   const Block & b = *_block;
   if( b.width() < 9 || b.height() > 3 * b.width() ) return 0;
 
-  const Blockmap & bm = *_blockmap;
-  const int cid = bm.id( b.vcenter(), b.hcenter() );
+  const bool cid = b.get_bit( b.vcenter(), b.hcenter() );
   const int noise = ( std::min( b.height(), b.width() ) / 15 ) + 1;
   int lrow1 = 0, urow2 = 0, lrow2 = 0, urow3 = 0;
   int lcol1 = 0, ucol2 = 0, lcol2 = 0, ucol3 = 0, black_section = 0;
@@ -52,14 +52,14 @@ int Features::test_235Esz( const Charset & charset ) const throw()
       {
       const int l = std::min( ucol, lcol ), r = std::max( ucol, lcol );
       int j;
-      for( j = l; j <= r; ++j ) if( bm.id( b.vcenter(), j ) != cid ) break;
+      for( j = l; j <= r; ++j ) if( b.get_bit( b.vcenter(), j ) != cid ) break;
       if( j <= r ) continue;
       }
     bool prev_black = false;
     for( int row = b.top(); row <= b.bottom(); ++row )
       {
       int col = ( row <= b.vcenter() ) ? ucol : lcol;
-      bool black = ( bm.id( row, col ) == b.id() );
+      bool black = b.get_bit( row, col );
       if( black )
         { if( !prev_black )
           { if( ++black_section == 2 ) { urow2 = row - 1; ucol2 = col; }
@@ -75,12 +75,13 @@ int Features::test_235Esz( const Charset & charset ) const throw()
     }
 
   if( black_section != 3 ) return 0;
+  bool tbopen = b.escape_bottom( urow3, ucol3 ) && b.escape_top( lrow1, lcol1 );
+  const int ascode = cid ? '*' : 0;
   if( b.escape_left( lrow2, lcol2 ) )
     {
     if( b.escape_left( urow2, ucol2 ) )
       {
-      if( b.escape_bottom( urow3, ucol3 ) && b.escape_top( lrow1, lcol1 ) )
-        { if( cid == b.id() ) return '*'; else return 0; }
+      if( tbopen ) return ascode;
       if( b.height() <= 3 * b.width() )
         {
         int lm = lp.minima(), rm = rp.minima();
@@ -91,6 +92,7 @@ int Features::test_235Esz( const Charset & charset ) const throw()
       }
     else if( b.escape_right( urow2, ucol2 ) )
       {
+      if( tbopen ) return ascode;
       if( b.seek_right( lrow1 + 1, lcol1 ) >= b.right() &&
           b.seek_left( urow3 - 1, ucol3 ) > b.left() )
         {
@@ -111,10 +113,12 @@ int Features::test_235Esz( const Charset & charset ) const throw()
     {
     if( b.escape_right( urow2, ucol2 ) )
       {
-      if( b.escape_bottom( urow3, ucol3 ) && b.escape_top( lrow1, lcol1 ) )
-        { if( cid == b.id() ) return '*'; else return 0; }
+      if( tbopen ) return ascode;
       if( bp.minima( b.height() / 5 ) == 1 )
         {
+        if( 8 * lp[((lrow2+urow3)/2)-b.top()] >= b.width() &&
+            b.escape_top( ( lrow1 + urow2 ) / 2, b.left() ) &&
+            !b.escape_top( ( lrow2 + urow3 ) / 2, b.left() ) ) return 'f';
         if( b.escape_bottom( urow3, ucol3 ) )
           {
           if( charset.enabled( Charset::iso_8859_15 ) ||
@@ -124,16 +128,14 @@ int Features::test_235Esz( const Charset & charset ) const throw()
               else return UCS::SCCEDI; }
           return 'F';
           }
-        if( 8 * lp[((lrow2+urow3)/2)-b.top()] >= b.width() &&
-            b.escape_top( ( lrow1 + urow2 ) / 2, b.left() ) &&
-            !b.escape_top( ( lrow2 + urow3 ) / 2, b.left() ) ) return 'f';
         else if( lrow1 < urow2 && lrow2 < urow3 ) return 'E';
         }
       }
     else if( b.escape_left( urow2, ucol2 ) )
       {
-      if( ( tp.isconvex() || ( tp.ispit() && !bp.ispit() ) ) &&
-          2 * lp[lp.pos(50)] >= b.width() ) return '2';
+      if( !tbopen && ( 2 * lp[lp.pos(50)] ) + 2 >= b.width() &&
+          ( tp.isconvex() || ( tp.ispit() && !bp.ispit() ) ) )
+        return '2';
       if( b.height() <= 2 * wp.max() && bp[bp.pos(75)] <= b.height() / 10 )
         return 'z';
       }
@@ -166,11 +168,11 @@ int Features::test_CEFIJLlT( const Charset & charset ) const throw()
 
   const int noise = ( std::min( b.height(), b.width() ) / 30 ) + 1;
   const int topmax = b.top() + noise;
-  if( vbars() == 1 && vbar(0).width() >= 2 &&
-      3 * vbar(0).width() <= b.width() + 2 )
+  const int botmin = b.bottom() - noise;
+  if( vbars() == 1 && vbar(0).width() >= 2 && 2 * vbar(0).width() < b.width() )
     {
-    if( std::abs( vbar(0).hcenter() - b.hcenter() ) <= 1 &&
-        std::abs( (vbar(0).left() - b.left()) - (b.right() - vbar(0).right()) ) <= 2 )
+    if( std::abs( vbar(0).hcenter() - b.hcenter() ) <= noise &&
+        std::abs( (vbar(0).left() - b.left()) - (b.right() - vbar(0).right()) ) <= 2 * noise )
       {
       if( hbars() == 1 && 4 * hbar(0).height() <= b.height() )
         {
@@ -179,8 +181,7 @@ int Features::test_CEFIJLlT( const Charset & charset ) const throw()
             Ocrad::similar( b.height(), b.width(), 50 ) ) return '+';
         }
       if( hbars() == 2 && 4 * hbar(0).height() <= b.height() &&
-          hbar(0).top() <= topmax &&
-          hbar(1).bottom() >= b.bottom() - 1 &&
+          hbar(0).top() <= topmax && hbar(1).bottom() >= botmin &&
           3 * hbar(0).width() > 4 * hbar(1).width() )
         return 'T';
       }
@@ -211,16 +212,22 @@ int Features::test_CEFIJLlT( const Charset & charset ) const throw()
             2 * hbar(1).width() >= 3 * vbar(0).width() &&
             hbar(1).includes_vcenter( b ) )
           return 'F';
-        if( hbar(1).bottom() >= b.bottom() - 1 &&
-            b.height() > b.width() &&
+        if( hbar(1).bottom() >= botmin && b.height() > b.width() &&
             hbar(1).width() > hbar(0).width() &&
             std::abs( vbar(0).hcenter() - hbar(0).hcenter() ) <= 1 &&
             rp.iminimum() > rp.pos( 70 ) )
           return 'L';
         }
-      if( hbars() == 1 && hbar(0).bottom() >= b.bottom() - 1 &&
-          Ocrad::similar( hbar(0).width(), b.width(), 10 ) )
-        return 'L';
+      if( hbars() == 1 && Ocrad::similar( hbar(0).width(), b.width(), 10 ) )
+        {
+        if( hbar(0).bottom() >= botmin &&
+            b.escape_top( b.vcenter(), b.hpos( 75 ) ) )
+          return 'L';
+        if( hbar(0).top() <= topmax && 2 * wp[wp.pos(50)] >= b.width() &&
+            4 * wp[wp.pos(75)] < b.width() &&
+            b.escape_right( b.vpos( 25 ), b.hcenter() ) )
+          return 'F';
+        }
       }
 
     if( vbar(0).left() > b.hcenter() &&
@@ -260,7 +267,7 @@ int Features::test_CEFIJLlT( const Charset & charset ) const throw()
       if( 2 * b.height() > 3 * b.width() )
         if( rp.isflats() || rp.minima() == 1 )
           if( vbar(0).right() >= b.hpos( 70 ) ||
-              b.escape_top( b.vpos( 75 ), vbar(0).right() + 1 ) )
+              b.escape_top( b.vpos( 75 ), std::min( b.right(), vbar(0).right() + 1 ) ) )
             for( int i = vbar(0).left() - 1; i > b.left(); --i )
               if( b.seek_bottom( b.vpos( 75 ), i ) < b.bottom() &&
                   bp[i-b.left()] <= noise ) return 'l';
@@ -293,7 +300,7 @@ int Features::test_CEFIJLlT( const Charset & charset ) const throw()
           vbar(0).height() >= 3 * vbar(0).width() &&
           lp.istpit() && lp.minima() == 1 )
         { int i = lp.iminimum();
-          if( i > lp.pos( 20 ) && i < lp.pos( 40 ) ) return '1'; }
+          if( i > lp.pos( 10 ) && i < lp.pos( 40 ) ) return '1'; }
       }
     }
   if( hbars() == 1 && std::abs( hbar(0).vcenter() - b.vcenter() ) <= 1 &&
@@ -376,10 +383,10 @@ int Features::test_frst() const throw()
   if( vbars() != 1 || vbar(0).width() < 2 ) return 0;
   if( vbar(0).hcenter() <= b.hcenter() )
     {
-    int col = b.right() - rp[b.vcenter()-b.top()] + 2;
+    const int col = b.right() - rp[b.vcenter()-b.top()] + 2;
     if( col < b.right() )
       {
-      int row = b.seek_bottom( b.vcenter(), col );
+      const int row = b.seek_bottom( b.vcenter(), col );
       if( row >= b.bottom() || b.escape_bottom( row - 1, col ) )
         {
         if( rp.minima() == 3 ) return 'f';
@@ -415,7 +422,7 @@ int Features::test_frst() const throw()
     if( 3 * b.height() >= 5 * b.width() && vbar(0).left() > b.left() &&
              rp.minima() <= 2 )
       {
-      int col = b.right() - rp[b.vcenter() - b.top()] + 1;
+      const int col = b.right() - std::max( 0, rp[b.vcenter() - b.top()] - 1 );
       if( !b.escape_bottom( b.vcenter(), col ) )
         { if( tp.ispit() && lp.iminimum() < lp.pos( 40 ) ) return 't'; }
       else if( 2 * wp.max() > b.width() )
@@ -434,6 +441,10 @@ int Features::test_frst() const throw()
       }
     if( b.seek_bottom( b.vcenter(), b.hpos( 60 ) + 1 ) >= b.bottom() )
       { if( rp.minima() == 2 ) return 'f'; else return 'r'; }
+    if( vbar(0).right() <= b.hcenter() && hbars() == 1 &&
+        hbar(0).bottom() >= b.bottom() - 1 && lp.istip() && rp.istip() &&
+        !b.escape_top( b.vcenter(), b.hpos( 75 ) ) )
+      return 'r';
     }
   return 0;
   }
@@ -477,6 +488,7 @@ int Features::test_G() const throw()
 
 
 // Common feature: U-shaped top of character
+//
 int Features::test_HKMNUuvwYy( const Rectangle & charbox ) const throw()
   {
   const Block & b = *_block;
@@ -514,7 +526,8 @@ int Features::test_HKMNUuvwYy( const Rectangle & charbox ) const throw()
         if( ( Ocrad::similar( lg, rg, 20 ) &&
               5 * wp[wp.pos(98)] >= 4 * wp[wp.pos(70)] ) ||
             ( lg > 1 && lg < rg && lc >= lc2 && !rp.increasing() ) ) return 'Y';
-      if( b.escape_top( b.vpos( 60 ), b.hcenter() ) ) return 'u';
+      if( b.escape_top( b.vpos( 60 ), b.hcenter() ) && !lp.istip() )
+        return 'u';
       if( lg < rg + 1 && !lp.increasing( lp.pos( 50 ) ) &&
           ( 2 * lg < rg || b.vpos( 90 ) >= charbox.bottom() ) &&
           ( tp.minima( b.height()/2 ) == 1 || lp.imaximum() > b.height()/2 ) )
@@ -534,7 +547,7 @@ int Features::test_HKMNUuvwYy( const Rectangle & charbox ) const throw()
       const int lg = lp[lp.pos(50)];
       const int rg = rp[rp.pos(50)];
       if( Ocrad::similar( lg, rg, 80, 2 ) &&
-          lg < b.width() / 4 && rg < b.width() / 4 )
+          4 * lg < b.width() && 4 * rg < b.width() )
         {
         if( lg > 1 && rg > 1 && lp.increasing() && rp.increasing() )
           return 'w';
@@ -545,6 +558,7 @@ int Features::test_HKMNUuvwYy( const Rectangle & charbox ) const throw()
         if( vscan()[rp.pos(60)] == 4 || vscan()[rp.pos(70)] == 4 ) return 'w';
         if( ( vbars() <= 2 || ( vbars() == 3 && b.height() >= b.width() ) ) &&
             tp.minima( b.height() / 2 ) <= 2 &&
+            tp.minima( ( 2 * b.height() ) / 5 ) <= 2 && !lp.istpit() &&
             4 * std::abs( rp[rp.pos(20)] - rp[rp.pos(80)] ) <= b.width() )
           return 'N';
         return 0;
@@ -562,6 +576,7 @@ int Features::test_HKMNUuvwYy( const Rectangle & charbox ) const throw()
 
 // Looks for the nearest frontier in column hcenter(), then tests if
 // gap is open downwards (except for 'x')
+//
 int Features::test_hknwx() const throw()
   {
   const Block & b = *_block;
@@ -586,7 +601,8 @@ int Features::test_hknwx() const throw()
     {
     const int m5 = tp.minima( b.height() / 5 );
     if( ( m5 == 2 || m5 == 3 ) && tp.minima() >= 2 &&
-        rp[rp.pos(25)] <= b.width() / 4 ) return 'w';
+        rp[rp.pos(25)] <= b.width() / 4 &&
+        ( !lp.istpit() || rp.minima() == 1 ) ) return 'w';
     if( m5 == 1 ) { if( rp.isctip( 66 ) ) return 'k'; else return 'h'; }
     return 0;
     }
@@ -603,10 +619,10 @@ int Features::test_hknwx() const throw()
 
 // Looks for four black sections in column hcenter()  1, then tests if
 // upper gap is open to the right and lower gaps are open to the left
+//
 int Features::test_s_cedilla() const throw()
   {
   const Block & b = *_block;
-  const Blockmap & bm = *_blockmap;
   int urow2 = 0, urow3 = 0, urow4 = 0, col, black_section = 0;
 
   for( col = b.hcenter() - 1; col <= b.hcenter() + 1; ++col )
@@ -614,7 +630,7 @@ int Features::test_s_cedilla() const throw()
     bool prev_black = false;
     for( int row = b.top(); row <= b.bottom(); ++row )
       {
-      bool black = ( bm.id( row, col ) == b.id() );
+      bool black = b.get_bit( row, col );
       if( black && !prev_black )
         {
         if( ++black_section == 2 ) urow2 = row - 1;
@@ -639,21 +655,20 @@ bool Features::test_comma() const throw()
   {
   const Block & b = *_block;
 
-  if( b.blocks() || b.height() <= b.width() || b.height() > 3 * b.width() )
+  if( b.holes() || b.height() <= b.width() || b.height() > 3 * b.width() )
     return false;
 
   if( b.width() >= 3 && b.height() >= 3 )
     {
-    const Blockmap & bm = *_blockmap;
     int upper_area = 0;
     for( int row = b.top(); row < b.top() + b.width(); ++row )
       for( int col = b.left(); col <= b.right(); ++col )
-        if( bm.id( row, col ) == b.id() ) ++upper_area;
+        if( b.get_bit( row, col ) ) ++upper_area;
     if( upper_area < (b.width() - 2) * (b.width() - 2) ) return false;
     int count1 = 0, count2 = 0;
     for( int col = b.left(); col <= b.right(); ++col )
-      { if( bm.id( b.top() + 1, col ) == b.id() ) ++count1;
-      if( bm.id( b.bottom() - 1, col ) == b.id() ) ++count2; }
+      { if( b.get_bit( b.top() + 1, col ) ) ++count1;
+      if( b.get_bit( b.bottom() - 1, col ) ) ++count2; }
     if( count1 <= count2 ) return false;
     }
   return true;
@@ -681,6 +696,7 @@ int Features::test_easy( const Rectangle & charbox ) const throw()
 
 // Recognizes single line, non-rectangular characters without holes.
 // '/<>\^`
+//
 int Features::test_line( const Rectangle & charbox ) const throw()
   {
   const Block & b = *_block;
@@ -740,12 +756,16 @@ int Features::test_line( const Rectangle & charbox ) const throw()
 int Features::test_solid( const Rectangle & charbox ) const throw()
   {
   const Block & b = *_block;
-  if( b.blocks() ) return 0;
+  if( b.holes() ) return 0;
 
-  if( b.height() > 3 && ( tp.minima() != 1 || bp.minima() != 1 ) ) return 0;
-  if( b.width() > 3 && ( lp.minima() != 1 || rp.minima() != 1 ) ) return 0;
+  if( b.height() >= 5 && b.width() >= 5 )
+    {
+    if( 2 * b.height() > b.width() && ( tp.minima() != 1 || bp.minima() != 1 ) )
+      return 0;
+    if( b.height() < 2 * b.width() && ( lp.minima() != 1 || rp.minima() != 1 ) )
+      return 0;
+    }
 
-  const Blockmap & bm = *_blockmap;
   int inner_area, inner_size, porosity = 0;
 
   if( b.width() >= 3 && b.height() >= 3 )
@@ -756,8 +776,8 @@ int Features::test_solid( const Rectangle & charbox ) const throw()
       {
       int holes = 0;	// FIXME
       for( int col = b.left() + 1; col < b.right(); ++col )
-        { if( bm.id( row, col ) == b.id() ) ++inner_area; else ++holes; }
-      if( 5 * holes >= b.width() ) ++porosity;
+        { if( b.get_bit( row, col ) ) ++inner_area; else ++holes; }
+      if( 5 * holes >= b.width() ) porosity += ( 5 * holes ) / b.width();
       }
     if( inner_area * 100 < inner_size * 70 ) return 0;
     }
@@ -771,8 +791,8 @@ int Features::test_solid( const Rectangle & charbox ) const throw()
       int d = 0;
       for( int i = 0; i < n; ++i )
         {
-        if( bm.id( b.top() + i, b.left() + i ) == b.id() ) ++d;
-        if( bm.id( b.top() + i, b.right() - i ) == b.id() ) --d;
+        if( b.get_bit( b.top() + i, b.left() + i ) ) ++d;
+        if( b.get_bit( b.top() + i, b.right() - i ) ) --d;
         }
       if( 2 * std::abs( d ) >= n - 1 ) return 0;
       }
@@ -782,11 +802,12 @@ int Features::test_solid( const Rectangle & charbox ) const throw()
             100 * b.area_octagon() >= 95 * b.area() ) ) ) return '.';
     return 0;
     }
-  if( porosity || inner_area * 100 < inner_size * 85 ) return 0;
+  if( porosity > 1 || inner_area * 100 < inner_size * 85 ||
+      ( porosity && inner_area * 100 < inner_size * 95 ) ) return 0;
   if( b.width() > b.height() )
     {
     if( b.width() >= 6 * b.height() &&
-        b.top() - b.height() > charbox.vcenter() ) return '_';
+        charbox.bottom() - b.bottom() < b.top() - charbox.vcenter() ) return '_';
     return '-';
     }
 
