@@ -13,7 +13,7 @@
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
-    Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
 
 #include <algorithm>
@@ -22,19 +22,19 @@
 #include "common.h"
 #include "rectangle.h"
 #include "ucs.h"
+#include "bitmap.h"
 #include "block.h"
-#include "blockmap.h"
 #include "profile.h"
 #include "feats.h"
 
 
 // Tests if the lower half of character is open to the left, to the right,
 // and/or to the bottom
+//
 int Features::test_49ARegpq( const Rectangle & charbox ) const throw()
   {
   const Block & b = *_block;
-  const Block & h = b.block( 0 );			// hole
-  const Blockmap & bm = *_blockmap;
+  const Bitmap & h = b.hole( 0 );
 
   if( bp.minima( b.height() / 10 ) == 2 && bp.isctip() && tp.minima() == 1 )
     { if( tp.isvpit() || rp.decreasing() ) return 'A'; else return 'R'; }
@@ -73,8 +73,8 @@ int Features::test_49ARegpq( const Rectangle & charbox ) const throw()
     if( b.bottom_hook( &hdiff ) && hdiff > 0 )
       {
       if( h.bottom() < b.vcenter() && h.right() + 2 <= b.right() &&
-          ( bm.id( h.bottom() + 1, h.right() + 1 ) != b.id() ||
-            bm.id( h.bottom() + 1, h.right() + 2 ) != b.id() || rp.isctip() ) )
+          ( !b.get_bit( h.bottom() + 1, h.right() + 1 ) ||
+            !b.get_bit( h.bottom() + 1, h.right() + 2 ) || rp.isctip() ) )
         return 's';
       else return 'g';
       }
@@ -91,7 +91,7 @@ int Features::test_49ARegpq( const Rectangle & charbox ) const throw()
 int Features::test_4ADQao( const Charset & charset, const Rectangle & charbox ) const throw()
   {
   const Block & b = *_block;
-  const Block & h = b.block( 0 );			// hole
+  const Bitmap & h = b.hole( 0 );
   int left_delta = h.left() - b.left(), right_delta = b.right() - h.right();
   const int noise = ( std::min( b.height(), b.width() ) / 30 ) + 1;
 
@@ -114,11 +114,14 @@ int Features::test_4ADQao( const Charset & charset, const Rectangle & charbox )
       }
     if( b.width() > 3 * h.width() )
       {
-      int c = vscan()[b.vcenter() - b.top()];
+      int c = vscan()[h.vcenter()-b.top()];
       int m = bp.minima();
       if( c == 3 && h.top() < b.vcenter() && h.bottom() > b.vcenter() &&
           3 * h.height() >= b.height() && ( m == 3 || m == 2 ) && !lp.ispit() )
         return 'm';
+      if( c == 3 && left_delta > right_delta && lp.ispit() &&
+          hscan()[h.hcenter()-b.left()] == 4 )
+        return '@';
       if( c == 4 && Ocrad::similar( left_delta, right_delta, 40 ) && lp.ispit() )
         return '@';
       }
@@ -144,10 +147,11 @@ int Features::test_4ADQao( const Charset & charset, const Rectangle & charbox )
 
 // Tests if the upper half of character is open to the left, to the right,
 // and/or to the bottom
+//
 int Features::test_6abd( const Charset & charset ) const throw()
   {
   const Block & b = *_block;
-  const Block & h = b.block( 0 );			// hole
+  const Bitmap & h = b.hole( 0 );
 
   if( 3 * h.width() < b.width() &&
       ( bp.minima( b.height() / 4 ) != 1 || tp.isctip() ) ) return 0;
