@@ -1,5 +1,5 @@
 /*  GNU Ocrad - Optical Character Recognition program
-    Copyright (C) 2003, 2004 Antonio Diaz Diaz.
+    Copyright (C) 2003, 2004, 2005 Antonio Diaz Diaz.
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -13,7 +13,7 @@
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
-    Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
 
 #include <algorithm>
@@ -21,24 +21,23 @@
 #include <vector>
 #include "common.h"
 #include "rectangle.h"
+#include "bitmap.h"
 #include "block.h"
-#include "blockmap.h"
 #include "profile.h"
 
 
-Profile::Profile( const Block & b, Type t ) throw()
-  : _block( &b ), _blockmap( b.blockmap() ), type( t ), valid( false ),
+Profile::Profile( const Bitmap & b, Type t ) throw()
+  : _bitmap( &b ), type( t ), valid( false ),
     valid_max( false ), valid_min( false ), valid_mean( false ),
     valid_isconcave( false ), valid_isconvex( false ),
     valid_isflat( false ), valid_isflats( false ),
-    valid_ispit( false ), valid_isupit( false ), valid_isvpit( false ),
-    valid_istip( false ) {}
+    valid_ispit( false ), valid_istpit( false ), valid_isupit( false ),
+    valid_isvpit( false ), valid_istip( false ) {}
 
 
 void Profile::initialize() throw()
   {
-  const Block & b = *_block;
-  const Blockmap & bm = *_blockmap;
+  const Bitmap & b = *_bitmap;
 
   valid = true;
   switch( type )
@@ -48,7 +47,7 @@ void Profile::initialize() throw()
       for( int i = 0; i < samples(); ++i )
         {
         int j = b.left(), row = b.top() + i;
-        while( j <= b.right() && bm.id( row, j ) != b.id() ) ++j;
+        while( j <= b.right() && !b.get_bit( row, j ) ) ++j;
         data[i] = j - b.left();
         } break;
     case top :
@@ -56,7 +55,7 @@ void Profile::initialize() throw()
       for( int i = 0; i < samples(); ++i )
         {
         int j = b.top(), col = b.left() + i;
-        while( j <= b.bottom() && bm.id( j, col ) != b.id() ) ++j;
+        while( j <= b.bottom() && !b.get_bit( j, col ) ) ++j;
         data[i] = j - b.top();
         } break;
     case right :
@@ -64,7 +63,7 @@ void Profile::initialize() throw()
       for( int i = 0; i < samples(); ++i )
         {
         int j = b.right(), row = b.top() + i;
-        while( j >= b.left() && bm.id( row, j ) != b.id() ) --j;
+        while( j >= b.left() && !b.get_bit( row, j ) ) --j;
         data[i] = b.right() - j;
         } break;
     case bottom :
@@ -72,7 +71,7 @@ void Profile::initialize() throw()
       for( int i = 0; i < samples(); ++i )
         {
         int j = b.bottom(), col = b.left() + i;
-        while( j >= b.top() && bm.id( j, col ) != b.id() ) --j;
+        while( j >= b.top() && !b.get_bit( j, col ) ) --j;
         data[i] = b.bottom() - j;
         } break;
     case height :
@@ -80,8 +79,8 @@ void Profile::initialize() throw()
       for( int i = 0; i < samples(); ++i )
         {
         int u = b.top(), d = b.bottom(), col = b.left() + i;
-        while( u <= d && bm.id( u, col ) != b.id() ) ++u;
-        while( u <= d && bm.id( d, col ) != b.id() ) --d;
+        while( u <= d && !b.get_bit( u, col ) ) ++u;
+        while( u <= d && !b.get_bit( d, col ) ) --d;
         data[i] = ( u <= d ) ? d - u + 1 : 0;
         } break;
     case width :
@@ -89,8 +88,8 @@ void Profile::initialize() throw()
       for( int i = 0; i < samples(); ++i )
         {
         int l = b.left(), r = b.right(), row = b.top() + i;
-        while( l <= r && bm.id( row, l ) != b.id() ) ++l;
-        while( l <= r && bm.id( row, r ) != b.id() ) --r;
+        while( l <= r && !b.get_bit( row, l ) ) ++l;
+        while( l <= r && !b.get_bit( row, r ) ) --r;
         data[i] = ( l <= r ) ? r - l + 1 : 0;
         } break;
     }
@@ -122,6 +121,16 @@ int Profile::max() throw()
   }
 
 
+int Profile::max( int l, int r ) throw()
+  {
+  if( !valid ) initialize();
+  if( r < 0 ) r = samples() - 1;
+  int m = 0;
+  for( int i = l; i <= r; ++i ) if( data[i] > m ) m = data[i];
+  return m;
+  }
+
+
 int Profile::min() throw()
   {
   if( !valid_min )
@@ -134,6 +143,16 @@ int Profile::min() throw()
   }
 
 
+int Profile::min( int l, int r ) throw()
+  {
+  if( !valid ) initialize();
+  if( r < 0 ) r = samples() - 1;
+  int m = _limit;
+  for( int i = l; i <= r; ++i ) if( data[i] < m ) m = data[i];
+  return m;
+  }
+
+
 int Profile::operator[]( int i ) throw()
   {
   if( !valid ) initialize();
@@ -143,12 +162,22 @@ int Profile::operator[]( int i ) throw()
   }
 
 
+int Profile::area( int l, int r ) throw()
+  {
+  if( !valid ) initialize();
+  if( r < 0 ) r = samples() - 1;
+  int area = 0;
+  for( int i = l; i <= r; ++i ) area += data[i];
+  return area;
+  }
+
+
 bool Profile::increasing( int i ) throw()
   {
   if( !valid ) initialize();
-  if( i < 0 || i > samples() - 2 || data[i] >= data[samples()-1] )
+  if( i <= 0 || i > samples() - 2 || data[samples()-1] - data[i] < 2 )
     return false;
-  for( ++i; i < samples(); ++i ) if( data[i] < data[i-1] ) return false;
+  while( ++i < samples() ) if( data[i] < data[i-1] ) return false;
   return true;
   }
 
@@ -156,9 +185,9 @@ bool Profile::increasing( int i ) throw()
 bool Profile::decreasing( int i ) throw()
   {
   if( !valid ) initialize();
-  if( i < 0 || i > samples() - 2 || data[i] <= data[samples()-1] )
+  if( i < 0 || i >= samples() - 2 || data[samples()-1] - data[i] > -2 )
     return false;
-  for( ++i; i < samples(); ++i ) if( data[i] > data[i-1] ) return false;
+  while( ++i < samples() - 1 ) if( data[i] > data[i-1] ) return false;
   return true;
   }
 
@@ -195,32 +224,42 @@ bool Profile::isconvex() throw()
   if( !valid_isconvex )
     {
     valid_isconvex = true; _isconvex = false; if( !valid ) initialize();
+    if( samples() < 9 || _limit < 5 ) return _isconvex;
     int min = _limit, min_begin = 0, min_end = 0;
     int lmin = _limit, rmax = -_limit, l = 0, r = 0;
 
     for( int i = 1; i < samples(); ++i )
       {
       int d = data[i] - data[i-1];
-      if( d < lmin ) { lmin = d; l = i; }
+      if( d < lmin ) { lmin = d; l = i - 1; }
       if( d >= rmax ) { rmax = d; r = i; }
       if( data[i] <= min )
         { min_end = i; if( data[i] < min ) { min = data[i]; min_begin = i; } }
       }
     if( l >= r || l >= pos( 25 ) || r <= pos( 75 ) ) return _isconvex;
-    if( lmin >= 0 || rmax <= 0 ) return _isconvex;
-    if( min_end - min_begin > samples() / 2 ) return _isconvex;
-    if( l >= min_begin || r <= min_end ) return _isconvex;
-    int imin = ( min_begin + min_end ) / 2;
-    if( imin < pos( 25 ) || imin > pos( 80 ) ) return _isconvex;
+    if( lmin >= 0 || rmax <= 0 || data[l] < 2 || data[r] < 2 ||
+        3 * ( data[l] + data[r] ) <= std::min( _limit, samples() ) )
+      return _isconvex;
+    if( 3 * ( min_end - min_begin + 1 ) > 2 * samples() ) return _isconvex;
+    if( 2 * l >= min_begin || 2 * r <= min_end + samples() - 1 ) return _isconvex;
+    if( min_begin < pos( 10 ) || min_end > pos( 90 ) ) return _isconvex;
 
+    const int noise = ( std::min( samples(), _limit ) / 30 ) + 1;
     int dmax = -_limit;
-    for( int i = l; i <= r; ++i )
+    for( int i = l + 1; i <= r; ++i )
       {
-      if( i > min_begin && i < min_end )
-        { if( data[i] <= 1 ) continue; else return _isconvex; }
+      if( i >= min_begin && i <= min_end )
+        { if( data[i] <= noise ) continue; else return _isconvex; }
       int d = data[i] - data[i-1];
-      if( d > dmax ) dmax = d;
-      else if( d < dmax - 2 || ( d < dmax - 1 && d != 0 && dmax != 0 ) )
+      if( d == 0 ) continue;
+      if( d > dmax ) { if( std::abs( d ) <= noise ) ++dmax; else dmax = d; }
+      else if( d < dmax - noise ) return _isconvex;
+      }
+    if( 2 * ( min_end - min_begin + 1 ) < samples() )
+      {
+      int varea = ( min_begin - l + 1 ) * data[l] / 2;
+      varea += ( r - min_end + 1 ) * data[r] / 2;
+      if( this->area( l, min_begin - 1 ) + this->area( min_end + 1, r ) >= varea )
         return _isconvex;
       }
     _isconvex = true;
@@ -274,12 +313,17 @@ bool Profile::ispit() throw()
     {
     valid_ispit = true; _ispit = false; if( !valid ) initialize();
     if( samples() < 5 ) return _ispit;
+    const int nsize = samples() / 30;
+    for( int i = 0; i <= nsize; ++i )
+      if( data[i] < nsize + 2 - i || data[samples()-i-1] < nsize + 2 - i )
+        return _ispit;
 
+    const int dmin = min();
     int begin = 0, end = 0, i, ref;
     for( i = 0, ref = _limit; i < samples(); ++i )
       {
       int d = data[i];
-      if( d == min() ) { begin = i; break; }
+      if( d == dmin ) { begin = i; break; }
       if( d < ref ) ref = d; else if( d > ref + 1 ) return _ispit;
       }
     if( begin < 2 || begin > samples() - 3 ) return _ispit;
@@ -287,13 +331,13 @@ bool Profile::ispit() throw()
     for( i = samples() - 1, ref = _limit; i >= begin; --i )
       {
       int d = data[i];
-      if( d == min() ) { end = i; break; }
+      if( d == dmin ) { end = i; break; }
       if( d < ref ) ref = d; else if( d > ref + 1 ) return _ispit;
       }
     if( end < begin || end > samples() - 3 ) return _ispit;
 
     for( i = begin + 1; i < end; ++i )
-      if( data[i] > min() + 1 ) return _ispit;
+      if( data[i] > dmin + 1 ) return _ispit;
 
     _ispit = true;
     }
@@ -325,6 +369,44 @@ bool Profile::iscpit( int cpos ) throw()
   }
 
 
+bool Profile::islpit() throw()
+  {
+  if( !valid ) initialize();
+  if( samples() < 5 ) return false;
+  const int nsize = samples() / 30;
+  if( data[0] < nsize + 2 ) return false;
+
+  const int dmin = min();
+  int begin = 0, ref = _limit;
+  for( int i = 0; i < samples(); ++i )
+    {
+    int d = data[i];
+    if( d == dmin ) { begin = i; break; }
+    if( d < ref ) ref = d; else if( d > ref + 1 ) return false;
+    }
+  if( begin < 2 || 2 * begin >= samples() ) return false;
+  return true;
+  }
+
+
+bool Profile::istpit() throw()
+  {
+  if( !valid_istpit )
+    {
+    valid_istpit = true; if( !valid ) initialize();
+    if( _limit < 5 || samples() < 5 || !ispit() )
+      { _istpit = false; return _istpit; }
+
+    const int noise = ( std::min( _limit, samples() ) / 30 ) + 1;
+    int l = -1, r = 0;
+    for( int i = 0; i < samples(); ++i )
+      if( data[i] <= noise ) { r = i; if( l < 0 ) l = i; }
+    _istpit = ( l > 0 && 4 * ( r - l + 1 ) < samples() );
+    }
+  return _istpit;
+  }
+
+
 bool Profile::isupit() throw()
   {
   if( !valid_isupit )
@@ -367,20 +449,20 @@ bool Profile::isvpit() throw()
   if( !valid_isvpit )
     {
     valid_isvpit = true; if( !valid ) initialize();
-    if( !ispit() ) { _isvpit = false; return _isvpit; }
+    if( _limit < 5 || samples() < 5 || !ispit() )
+      { _isvpit = false; return _isvpit; }
 
-    int l1 = -1, l2 = -1, r1 = -2, r2 = -2;
+    const int noise = _limit / 20;
+    const int level = ( _limit / 10 ) + 2;
+    int ll = -1, ln = -1, rl = -1, rn = -1;
     for( int i = 0; i < samples(); ++i )
-      if( data[i] <= 2 )
+      if( data[i] <= level )
         {
-        r2 = i; if( l2 < 0 ) l2 = i;
-        if( data[i] <= 1 ) { r1 = i; if( l1 < 0 ) l1 = i; }
+        rl = i; if( ll < 0 ) ll = i;
+        if( data[i] <= noise ) { rn = i; if( ln < 0 ) ln = i; }
         }
-    int count1 = r1 - l1 + 1, count2 = r2 - l2 + 1 - count1;
-    _isvpit = ( 4 * count1 < samples() ||
-                ( 3 * count1 <= samples() && count2 <= 2 ) ||
-                ( 5 * count1 <= 2 * samples() && count2 <= 1 ) ||
-                ( 2 * count1 <= samples() && count2 <= 0 ) );
+    const int wl = rl - ll + 1, wn = rn - ln + 1;
+    _isvpit = ( ln > 0 && 2 * wl <= samples() && wl - wn <= 2 * ( level - noise ) );
     }
   return _isvpit;
   }
@@ -403,6 +485,7 @@ bool Profile::istip() throw()
         { rth = data[samples()-1-i]; end = samples() - 1 - i; }
       }
     if( lth >= th || rth >= th ) return _istip;
+    if( 3 * lth >= 2 * range() || 3 * rth >= 2 * range() ) return _istip;
     th = std::max( lth, rth );
     int status = 0;
     for( int i = begin + 1; i < end; ++i )
@@ -423,7 +506,7 @@ bool Profile::isctip( int cpos ) throw()
   {
   if( !valid ) initialize();
   if( samples() < 5 || cpos < 25 || cpos > 75 ) return false;
-  int th = ( mean() < 2 ) ? 2 : mean();
+  int th = std::max( 2, mean() );
   int imax = -1, mid = ( ( samples() - 1 ) * cpos ) / 100;
 
   for( int i = 0; i < samples() / 4; ++i )
@@ -431,7 +514,16 @@ bool Profile::isctip( int cpos ) throw()
     if( data[mid+i] > th ) { imax = mid + i; break; }
     if( data[mid-i-1] > th ) { imax = mid - i - 1; break; }
     }
-  if( imax < 0 ) return false;
+  if( imax < 0 && mean() == 0 )
+    {
+    --th;
+    for( int i = 0; i < samples() / 4; ++i )
+      {
+      if( data[mid+i] > th ) { imax = mid + i; break; }
+      if( data[mid-i-1] > th ) { imax = mid - i - 1; break; }
+      }
+    if( imax < 0 ) return false;
+    }
 
   for( int i = imax + 1; i < samples(); ++i )
     if( data[i] < th )
@@ -446,9 +538,10 @@ bool Profile::isctip( int cpos ) throw()
 int Profile::imaximum() throw()
   {
   if( !valid ) initialize();
+  const int margin = ( samples() / 30 ) + 1;
   int mbegin = 0, mend, mvalue = 0;
 
-  for( int i = 0; i < samples(); ++i )
+  for( int i = margin; i < samples() - margin; ++i )
     if( data[i] > mvalue ) { mvalue = data[i]; mbegin = i; }
 
   for( mend = mbegin + 1; mend < samples(); ++mend )
@@ -472,8 +565,8 @@ int Profile::iminimum( int m, int th ) throw()
       { if( data[end] < th ) { status = 1; ++minima; begin = end; } }
     else if( data[end] > th )
       { if( minima == m + 1 ) { --end; break; } else status = 0; }
-    else if( end == samples() - 1 ) break;
     }
+  if( end >= samples() ) --end;
   if( minima != m + 1 ) return 0;
 
   for( int i = begin; i <= end; ++i )
@@ -483,29 +576,22 @@ int Profile::iminimum( int m, int th ) throw()
   }
 
 
-int Profile::maxima() throw()
-  {
-  if( !valid ) initialize();
-  int maxima = ( data[0] > mean() ) ? 1 : 0;
-  for( int i = 1; i < samples(); ++i )
-    if( data[i] > mean() && data[i-1] <= mean() ) ++maxima;
-  return maxima;
-  }
-
-
 int Profile::minima( int th ) throw()
   {
   if( !valid ) initialize();
   if( !samples() ) return false;
   if( th < 1 ) th = ( mean() < 2 ) ? 2 : mean();
-  int minima = ( data[0] < th ) ? 1 : 0;
+  const int noise = _limit / 40;
+  const int dth = th - ( ( noise + 1 ) / 2 ), uth = th + ( noise / 2 );
+  if( dth < 1 ) return 1;
+  int minima = ( data[0] < dth ) ? 1 : 0;
   int status = ( minima ) ? 1 : 0;
 
   for( int i = 1; i < samples(); ++i )
     switch( status )
       {
-      case 0: if( data[i] < th ) { status = 1; ++minima; } break;
-      case 1: if( data[i] > th ) status = 0; break;
+      case 0: if( data[i] < dth ) { status = 1; ++minima; } break;
+      case 1: if( data[i] > uth ) status = 0; break;
       }
   return minima;
   }
@@ -516,16 +602,19 @@ bool Profile::straight( int * _dy ) throw()
   if( !valid ) initialize();
   if( samples() < 5 ) return false;
 
+  const int xl = ( samples() / 30 ) + 1, yl = ( data[xl] + data[xl+1] ) / 2 ;
+  const int xr = samples() - xl - 1,     yr = ( data[xr-1] + data[xr] ) / 2 ;
+  const int dx = xr - xl, dy = yr - yl;
+  if( dx <= 0 ) return false;
+  const int dmax = dx * ( ( samples() / 20 ) + 2 );
   int faults = samples() / 10;
-  int xl = ( samples() / 20 ) + 1, yl = data[xl];
-  int xr = samples() - xl - 1,     yr = data[xr];
-  int dx = xr - xl, dy = yr - yl;
-  for( int i = xl + 1; i < xr; ++i )
+  for( int i = 0; i < samples(); ++i )
     {
-    int y = yl; if( dx ) y += ( dy * ( i - xl ) / dx );
-    int d = std::abs( y - data[i] );
-    if( d > 2 || ( d > 1 && --faults <= 0 ) ) return false;
+    int y = ( dx * yl ) + ( ( i - xl ) * dy );
+    int d = std::abs( ( dx * data[i] ) - y );
+    if( d >= dmax && ( ( dx * data[i] ) < y || ( i >= xl && i <= xr ) ) )
+      if( d > dmax || ( d == dmax && --faults < 0 ) ) return false;
     }
-  *_dy = dy;
+  if( _dy ) *_dy = dy;
   return true;
   }
