@@ -13,7 +13,7 @@
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
-    Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
 
 #include <cstdio>
@@ -31,24 +31,28 @@
 
 namespace {
 
-void add_line( std::vector< Textline > & textline_vector ) throw()
-  { textline_vector.push_back( Textline() ); }
+void add_line( std::vector< Textline * > & textlinep_vector ) throw()
+  { textlinep_vector.push_back( new Textline() ); }
 
 
-void insert_line( std::vector< Textline > & textline_vector, int i ) throw()
-  { textline_vector.insert( textline_vector.begin() + i, Textline() ); }
+void insert_line( std::vector< Textline * > & textlinep_vector, int i ) throw()
+  { textlinep_vector.insert( textlinep_vector.begin() + i, new Textline() ); }
 
 
-void delete_line( std::vector< Textline > & textline_vector, int i ) throw()
-  { textline_vector.erase( textline_vector.begin() + i ); }
+void delete_line( std::vector< Textline * > & textlinep_vector, int i ) throw()
+  {
+  delete textlinep_vector[i];
+  textlinep_vector.erase( textlinep_vector.begin() + i );
+  }
 
 
 // Build the vertical composite characters.
-void join_characters( std::vector< Textline > & textline_vector ) throw()
+//
+void join_characters( std::vector< Textline * > & tlpv ) throw()
   {
-  for( int current_line = 0; current_line < (int)textline_vector.size(); ++current_line )
+  for( unsigned int current_line = 0; current_line < tlpv.size(); ++current_line )
     {
-    Textline & line = textline_vector[current_line];
+    Textline & line = *tlpv[current_line];
     for( int i = 0 ; i < line.characters() - 1; )
       {
       Character & c1 = line.character( i );
@@ -81,32 +85,34 @@ void join_characters( std::vector< Textline > & textline_vector ) throw()
 
 
 Textblock::Textblock( const Rectangle & r, const Charset & charset,
-           const std::vector< std::vector< Block > > & block_matrix ) throw()
+           std::vector< std::vector< Block * > > & blockp_matrix ) throw()
   : Rectangle( r )
   {
-  int cuts = block_matrix.size();
+  int cuts = blockp_matrix.size();
   std::vector< int > mean_height( cuts, 0 );
-  std::vector< std::vector< Block > > pending( cuts ), pending_tall( cuts ), pending_short( cuts );
+  std::vector< std::vector< Block * > > pending( cuts );
+  std::vector< std::vector< Block * > > pending_tall( cuts );
+  std::vector< std::vector< Block * > > pending_short( cuts );
 
   // Classify blocks by height.
   for( int cut = 0; cut < cuts; ++cut )
     {
-    const std::vector< Block > & block_vector = block_matrix[cut];
-    if( !block_vector.size() ) continue;
+    const std::vector< Block * > & blockp_vector = blockp_matrix[cut];
+    if( !blockp_vector.size() ) continue;
     unsigned int samples = 0;
     std::vector< int > height_distrib;
-    for( unsigned int i = 0; i < block_vector.size(); ++i )
+    for( unsigned int i = 0; i < blockp_vector.size(); ++i )
       {
-      unsigned int h = block_vector[i].height();
-      unsigned int w = block_vector[i].width();
+      unsigned int h = blockp_vector[i]->height();
+      unsigned int w = blockp_vector[i]->width();
       if( h < 10 || w >= 3 * h ) continue;
       if( h >= height_distrib.size() ) height_distrib.resize( h + 1 );
       ++height_distrib[h]; ++samples;
       }
     if( !height_distrib.size() )
-      for( unsigned int i = 0; i < block_vector.size(); ++i )
+      for( unsigned int i = 0; i < blockp_vector.size(); ++i )
         {
-        unsigned int h = block_vector[i].height();
+        unsigned int h = blockp_vector[i]->height();
         if( h >= height_distrib.size() ) height_distrib.resize( h + 1 );
         ++height_distrib[h]; ++samples;
         }
@@ -120,66 +126,66 @@ Textblock::Textblock( const Rectangle & r, const Charset & charset,
       }
     if( valid_samples ) mean_height[cut] /= valid_samples;
 
-    for( unsigned int i = 0; i < block_vector.size(); ++i )
+    for( unsigned int i = 0; i < blockp_vector.size(); ++i )
       {
-      const Block & b = block_vector[i];
-      if( b.height() >= 2 * mean_height[cut] )
-        pending_tall[cut].push_back( b );
-      else if( 2 * b.height() <= mean_height[cut] )
-        pending_short[cut].push_back( b );
-      else pending[cut].push_back( b );
+      Block * p = blockp_vector[i];
+      if( p->height() >= 2 * mean_height[cut] )
+        pending_tall[cut].push_back( p );
+      else if( 2 * p->height() <= mean_height[cut] )
+        pending_short[cut].push_back( p );
+      else pending[cut].push_back( p );
       }
     }
 
   // Assign normal blocks to characters and create lines.
   int min_line = 0;	// first line of current cut
-  add_line( data );
+  add_line( tlpv );
   for( int cut = 0; cut < cuts; ++cut )
     {
     if( pending[cut].size() )
       {
-      if( line( lines() - 1 ).characters() ) add_line( data );
-      int current_line = min_line = lines() - 1;
-      line( current_line ).shift_character( Character( pending[cut][0] ) );
+      if( tlpv.back()->characters() ) add_line( tlpv );
+      int current_line = min_line = textlines() - 1;
+      tlpv[current_line]->shift_characterp( new Character( pending[cut][0] ) );
       for( unsigned int i = 1; i < pending[cut].size(); ++i )
         {
-        Character c( pending[cut][i] );
+        Block & b = *pending[cut][i];
         current_line = std::max( min_line, current_line - 2 );
         while( true )
           {
           const Character *cl = 0, *cr = 0;
-          for( int j = line( current_line ).characters() - 1; j >= 0; --j )
+          for( int j = tlpv[current_line]->characters() - 1; j >= 0; --j )
             {
-            const Character & cj = line( current_line ).character( j );
-            if( !c.includes_hcenter( cj ) && !cj.includes_hcenter( c ) )
-              { if( c.h_precedes( cj ) ) cr = &cj; else { cl = &cj; break; } }
+            const Character & cj = tlpv[current_line]->character( j );
+            if( !b.includes_hcenter( cj ) && !cj.includes_hcenter( b ) )
+              { if( b.h_precedes( cj ) ) cr = &cj; else { cl = &cj; break; } }
             }
-          if( ( cl && ( cl->includes_vcenter( c ) || c.includes_vcenter( *cl ) ) ) ||
-              ( cr && ( cr->includes_vcenter( c ) || c.includes_vcenter( *cr ) ) ) )
-            { line( current_line ).shift_character( c ); break; }
-          else if( ( cl && cl->bottom() < c.top() ) || ( cr && cr->bottom() < c.top() ) )
+          if( ( cl && ( cl->includes_vcenter( b ) || b.includes_vcenter( *cl ) ) ) ||
+              ( cr && ( cr->includes_vcenter( b ) || b.includes_vcenter( *cr ) ) ) )
+            { tlpv[current_line]->shift_characterp( new Character( &b ) ); break; }
+          else if( ( cl && cl->bottom() < b.top() ) || ( cr && cr->bottom() < b.top() ) )
             {
-            if( ++current_line >= lines() )
-              { add_line( data ); current_line = lines() - 1;
-              line( current_line ).shift_character( c ); break; }
+            if( ++current_line >= textlines() )
+              { add_line( tlpv ); current_line = textlines() - 1;
+              tlpv[current_line]->shift_characterp( new Character( &b ) ); break; }
             }
-          else if( ( cl && cl->top() > c.bottom() ) || ( cr && cr->top() > c.bottom() ) )
+          else if( ( cl && cl->top() > b.bottom() ) || ( cr && cr->top() > b.bottom() ) )
             {
-            insert_line( data, current_line );
-            line( current_line ).shift_character( c ); break;
+            insert_line( tlpv, current_line );
+            tlpv[current_line]->shift_characterp( new Character( &b ) ); break;
             }
-          else if( ( cl && cl->v_overlaps( c ) ) || ( cr && cr->v_overlaps( c ) ) )
-            { line( current_line ).shift_character( c ); break; }
-          else break;
+          else if( ( cl && cl->v_overlaps( b ) ) || ( cr && cr->v_overlaps( b ) ) )
+            { tlpv[current_line]->shift_characterp( new Character( &b ) ); break; }
+          else { delete &b; break; }
           }
         }
       }
     }
 
-  join_characters( data );
+  join_characters( tlpv );
 
   // Create tracks of lines.
-  for( int i = 0; i < lines(); ++i ) line( i ).set_track();
+  for( int i = 0; i < textlines(); ++i ) tlpv[i]->set_track();
 
   // Insert tall blocks.
   // Seek up, then seek down, needed for slanted or curved lines.
@@ -187,18 +193,19 @@ Textblock::Textblock( const Rectangle & r, const Charset & charset,
     {
     for( unsigned int i = 0; i < pending_tall[cut].size(); ++i )
       {
-      Character c( pending_tall[cut][i] );
+      Block & b = *pending_tall[cut][i];
       while( current_line > 0 &&
-             c.bottom() < line( current_line ).vcenter( c.hcenter() ) )
+             b.bottom() < tlpv[current_line]->vcenter( b.hcenter() ) )
         --current_line;
-      while( current_line < lines() &&
-             c.top() > line( current_line ).vcenter( c.hcenter() ) )
+      while( current_line < textlines() &&
+             b.top() > tlpv[current_line]->vcenter( b.hcenter() ) )
         ++current_line;
-      if( current_line >= lines() ) { --current_line; continue; }
-      if( c.height() > 2 * line( current_line ).mean_height() )
-        { if( c.hcenter() < line( current_line ).character(0).left() )
-            line( current_line ).big_initial( c ); }
-      else line( current_line ).shift_character( c );
+      if( current_line >= textlines() )
+        { --current_line; delete &b; continue; }
+      if( b.height() > 2 * tlpv[current_line]->mean_height() )
+        { if( b.hcenter() < tlpv[current_line]->character(0).left() )
+            tlpv[current_line]->big_initial( new Character( &b ) ); }
+      else tlpv[current_line]->shift_characterp( new Character( &b ) );
       }
     }
 
@@ -208,35 +215,35 @@ Textblock::Textblock( const Rectangle & r, const Charset & charset,
     {
     for( unsigned int i = 0; i < pending_short[cut].size(); ++i )
       {
-      Character c( pending_short[cut][i] );
+      Block & b = *pending_short[cut][i];
       while( current_line > 0 &&
-             c.bottom() < line( current_line ).top( c.hcenter() ) )
+             b.bottom() < tlpv[current_line]->top( b.hcenter() ) )
         --current_line;
       int temp = current_line;
-      while( current_line < lines() &&
-             c.top() > line( current_line ).bottom( c.hcenter() ) )
+      while( current_line < textlines() &&
+             b.top() > tlpv[current_line]->bottom( b.hcenter() ) )
         ++current_line;
-      if( current_line >= lines() )
+      if( current_line >= textlines() )
         {
-        const Textline & l = line( --current_line );
-        if( c.top() > l.bottom( c.hcenter() ) + l.height() ) continue;
+        const Textline & l = *tlpv[--current_line];
+        if( b.top() > l.bottom( b.hcenter() ) + l.height() ) continue;
         else temp = current_line;
         }
       if( current_line - temp > 1 ) temp = current_line - 1;
       if( current_line != temp &&
-          2 * ( c.top() - line( temp ).bottom( c.hcenter() ) ) <
-          line( current_line ).top( c.hcenter() ) - c.bottom() )
+          2 * ( b.top() - tlpv[temp]->bottom( b.hcenter() ) ) <
+          tlpv[current_line]->top( b.hcenter() ) - b.bottom() )
         current_line = temp;
-      line( current_line ).shift_character( c );
+      tlpv[current_line]->shift_characterp( new Character( &b ) );
       }
     }
 
   // Second pass. Join lines of i-dots and tildes.
-  for( int current_line = 0; current_line < lines() - 1; )
+  for( int current_line = 0; current_line < textlines() - 1; )
     {
     bool joined = false;
-    Textline & line1 = line( current_line );
-    Textline & line2 = line( current_line + 1 );
+    Textline & line1 = *tlpv[current_line];
+    Textline & line2 = *tlpv[current_line+1];
     if( line1.characters() <= 2 * line2.characters() &&
         2 * line1.mean_height() < line2.mean_height() )
       for( int i1 = 0; !joined && i1 < line1.characters(); ++i1 )
@@ -252,35 +259,35 @@ Textblock::Textblock( const Rectangle & r, const Charset & charset,
               && c2.top() - c1.bottom() < line2.mean_height() )
             {
             joined = true; line2.join( line1 );
-            delete_line( data, current_line );
+            delete_line( tlpv, current_line );
             }
           }
         }
     if( !joined ) ++current_line;
     }
 
-  join_characters( data );
+  join_characters( tlpv );
 
   // Fourth pass. Remove noise lines.
-  if( lines() >= 3 )
+  if( textlines() >= 3 )
     {
-    for( int i = 0; i + 2 < lines(); ++i )
+    for( int i = 0; i + 2 < textlines(); ++i )
       {
-      Textline & line1 = line( i );
-      Textline & line2 = line( i + 1 );
-      Textline & line3 = line( i + 2 );
+      Textline & line1 = *tlpv[i];
+      Textline & line2 = *tlpv[i+1];
+      Textline & line3 = *tlpv[i+2];
       if( line2.characters() > 2 || line1.characters() < 4 ||
           line3.characters() < 4 ) continue;
       if( !Ocrad::similar( line1.height(), line3.height(), 10 ) ) continue;
       if( 8 * line2.height() > line1.height() + line3.height() ) continue;
-      delete_line( data, i + 1 );
+      delete_line( tlpv, i + 1 );
       }
     }
 
   // Fifth pass. Add spaces between characters.
-  for( int current_line = 0; current_line < lines(); ++current_line )
+  for( int current_line = 0; current_line < textlines(); ++current_line )
     {
-    Textline & line1 = line( current_line );
+    Textline & line1 = *tlpv[current_line];
     const int mw = line1.mean_width();
     const int space_width_limit = line1.space_width_limit( mw );
     const int mg = line1.mean_gap_width( space_width_limit );
@@ -309,64 +316,72 @@ Textblock::Textblock( const Rectangle & r, const Charset & charset,
     }
 
   // Recognize characters.
-  for( int i = 0; i < lines(); ++i )
+  for( int i = 0; i < textlines(); ++i )
     {
     // First pass. Recognize the easy characters.
-    line( i ).recognize1( charset );
+    tlpv[i]->recognize1( charset );
     // Second pass. Use context to clear up ambiguities.
-    line( i ).recognize2( charset );
+    tlpv[i]->recognize2( charset );
     }
 
   // Remove unrecognized lines.
-  for( int i = lines() - 1; i >= 0; --i )
+  for( int i = textlines() - 1; i >= 0; --i )
     {
-    Textline & line1 = line( i );
+    Textline & line1 = *tlpv[i];
     bool flag = false;
     if( line1.big_initial() && line1.big_initial()->guesses() ) flag = true;
     else
       for( int j = 0 ; j < line1.characters(); ++j )
         { if( line1.character( j ).guesses() ) { flag = true; break; } }
-    if( !flag ) delete_line( data, i );
+    if( !flag ) delete_line( tlpv, i );
     }
 
   // Add blank lines.
-  if( lines() >= 3 )
+  if( textlines() >= 3 )
     {
-    int mean_vdistance = ( line( lines() - 1 ).mean_vcenter() - line( 0 ).mean_vcenter() ) / ( lines() - 1 );
+    int mean_vdistance = ( tlpv.back()->mean_vcenter() - tlpv.front()->mean_vcenter() ) / ( textlines() - 1 );
     if( mean_vdistance > 0 )
-      for( int current_line = 0; current_line + 1 < lines(); ++current_line )
+      for( int current_line = 0; current_line + 1 < textlines(); ++current_line )
         {
-        Textline & line1 = line( current_line );
-        Textline & line2 = line( current_line + 1 );
+        Textline & line1 = *tlpv[current_line];
+        Textline & line2 = *tlpv[current_line+1];
         int vdistance = line2.mean_vcenter() - line1.mean_vcenter();
         if( vdistance <= 0 ) continue;
         int newlines = (vdistance + (mean_vdistance / 2)) / mean_vdistance;
-        for( int i = 1; i < newlines; ++i ) insert_line( data, ++current_line );
+        for( int i = 1; i < newlines; ++i ) insert_line( tlpv, ++current_line );
         }
     }
 
   }
 
 
+Textblock::~Textblock() throw()
+  {
+  for( int i = textlines() - 1; i >= 0; --i ) delete tlpv[i];
+  }
+
+
 int Textblock::characters() const throw()
   {
   int total = 0;
-  for( int i = 0; i < lines(); ++i ) total += line( i ).characters();
+  for( int i = 0; i < textlines(); ++i )
+    total += tlpv[i]->characters();
   return total;
   }
 
-
-Textline & Textblock::line( int i ) const throw()
+/*
+Textline & Textblock::textline( int i ) const throw()
   {
-  if( i < 0 || i >= lines() )
+  if( i < 0 || i >= textlines() )
     Ocrad::internal_error( "line, index out of bounds" );
-  return data[i];
+  return *tlpv[i];
   }
-
+*/
 
 void Textblock::print( const Control & control ) const throw()
   {
-  for( int i = 0; i < lines(); ++i ) line( i ).print( control );
+  for( int i = 0; i < textlines(); ++i )
+    tlpv[i]->print( control );
   std::fputs( "\n", control.outfile );
   }
 
@@ -374,13 +389,13 @@ void Textblock::print( const Control & control ) const throw()
 void Textblock::dprint( const Control & control, bool graph, bool recursive )
 								const throw()
   {
-  std::fprintf( control.outfile, "%d lines\n\n", lines() );
+  std::fprintf( control.outfile, "%d lines\n\n", textlines() );
 
-  for( int i = 0; i < lines(); ++i )
+  for( int i = 0; i < textlines(); ++i )
     {
     std::fprintf( control.outfile, "%d characters in line %d\n",
-                  line( i ).characters(), i+1 );
-    line( i ).dprint( control, graph, recursive );
+                  tlpv[i]->characters(), i + 1 );
+    tlpv[i]->dprint( control, graph, recursive );
     }
   std::fputs( "\n", control.outfile );
   }
@@ -388,24 +403,24 @@ void Textblock::dprint( const Control & control, bool graph, bool recursive )
 
 void Textblock::xprint( const Control & control ) const throw()
   {
-  std::fprintf( control.exportfile, "lines %d\n", lines() );
+  std::fprintf( control.exportfile, "lines %d\n", textlines() );
 
-  for( int i = 0; i < lines(); ++i )
+  for( int i = 0; i < textlines(); ++i )
     {
     std::fprintf( control.exportfile, "line %d chars %d height %d\n", i + 1,
-                  line( i ).characters(), line( i ).mean_height() );
-    line( i ).xprint( control );
+                  tlpv[i]->characters(), tlpv[i]->mean_height() );
+    tlpv[i]->xprint( control );
     }
   }
 
 
 void Textblock::cmark( Bitmap & bitmap ) const throw()
   {
-  for( int i = 0; i < lines(); ++i ) line( i ).cmark( bitmap );
+  for( int i = 0; i < textlines(); ++i ) tlpv[i]->cmark( bitmap );
   }
 
 
 void Textblock::lmark( Bitmap & bitmap ) const throw()
   {
-  for( int i = 0; i < lines(); ++i ) bitmap.draw_track( line( i ) );
+  for( int i = 0; i < textlines(); ++i ) bitmap.draw_track( *tlpv[i] );
   }
