@@ -1,5 +1,5 @@
 /*  GNU Ocrad - Optical Character Recognition program
-    Copyright (C) 2003, 2004 Antonio Diaz Diaz.
+    Copyright (C) 2003, 2004, 2005 Antonio Diaz Diaz.
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -13,7 +13,7 @@
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
-    Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
 
 #include <cctype>
@@ -31,10 +31,11 @@
 
 
 Textline::Textline( const Textline & tl ) throw()
-  : Track( tl ), data( tl.data )
+  : Track( tl )
   {
-  if( tl._big_initial ) _big_initial = new Character( *(tl._big_initial) );
-  else _big_initial = 0;
+  big_initial( tl._big_initial );
+  for( unsigned int i = 0; i < tl.cpv.size(); ++i )
+    shift_characterp( new Character( *tl.cpv[i] ) );
   }
 
 
@@ -43,50 +44,59 @@ Textline & Textline::operator=( const Textline & tl ) throw()
   if( this != &tl )
     {
     Track::operator=( tl );
-    delete_big_initial();
-    if( tl._big_initial ) _big_initial = new Character( *(tl._big_initial) );
-    data = tl.data;
+    big_initial( tl._big_initial );
+    for( unsigned int i = 0; i < cpv.size(); ++i ) delete cpv[i];
+    cpv.clear();
+    for( unsigned int i = 0; i < tl.cpv.size(); ++i )
+      shift_characterp( new Character( *tl.cpv[i] ) );
     }
   return *this;
   }
 
 
+Textline::~Textline() throw()
+  {
+  big_initial( 0 );
+  for( unsigned int i = 0; i < cpv.size(); ++i ) delete cpv[i];
+  }
+
+
 void Textline::set_track() throw()
   {
   std::vector< Rectangle > rv;
-  for( unsigned int i = 0; i < data.size(); ++i )
-    if( !data[i].maybe(' ') ) rv.push_back( data[i] );
+  for( unsigned int i = 0; i < cpv.size(); ++i )
+    if( !cpv[i]->maybe(' ') ) rv.push_back( *cpv[i] );
   Track::set_track( rv );
   }
 
 
-void Textline::big_initial( const Character & c ) throw()
+void Textline::big_initial( Character * p ) throw()
   {
-  if( !_big_initial ) _big_initial = new Character( c );
-  else *_big_initial = c;
+  if( _big_initial ) delete _big_initial;
+  _big_initial = p;
   }
 
-
+/*
 void Textline::delete_big_initial() throw()
   {
   if( _big_initial ) { delete _big_initial; _big_initial = 0; }
   }
-
+*/
 
 void Textline::delete_character( int i ) throw()
   {
   if( i < 0 || i >= characters() )
     Ocrad::internal_error( "delete_character, index out of bounds" );
-  data.erase( data.begin() + i );
+  delete cpv[i]; cpv.erase( cpv.begin() + i );
   }
 
 
-int Textline::shift_character( const Character & c ) throw()
+int Textline::shift_characterp( Character * p ) throw()
   {
   int i = characters();
 
-  while( i > 0 && c.h_precedes( data[i-1] ) ) --i;
-  data.insert( data.begin() + i, c );
+  while( i > 0 && p->h_precedes( *cpv[i-1] ) ) --i;
+  cpv.insert( cpv.begin() + i, p );
   return i;
   }
 
@@ -97,26 +107,46 @@ bool Textline::insert_space( int i, bool tab ) throw()
     Ocrad::internal_error( "insert_space, index out of bounds" );
   if( height() == 0 )
     Ocrad::internal_error( "insert_space, track not set yet" );
-  Character & c1 = data[i-1];
-  Character & c2 = data[i];
+  Character & c1 = *cpv[i-1];
+  Character & c2 = *cpv[i];
   int l = c1.right() + 1;
   int r = c2.left() - 1;
   if( l > r ) return false;
   int t = top( ( l + r ) / 2 );
   int b = bottom( ( l + r ) / 2 );
-  Character c( Rectangle( l, t, r, b ), *c1.blockmap() );
-  c.add_guess( ' ', tab ? 1 : 0 );
-  if( tab ) c.add_guess( '\t', 0 );
-  data.insert( data.begin() + i, c );
+  Rectangle re( l, t, r, b );
+  Character * p = new Character( re, ' ', tab ? 1 : 0 );
+  if( tab ) p->add_guess( '\t', 0 );
+  cpv.insert( cpv.begin() + i, p );
   return true;
   }
 
 
+void Textline::join( Textline & tl ) throw()
+  {
+  if( tl._big_initial &&
+      ( !_big_initial || tl._big_initial->h_precedes( *_big_initial ) ) )
+    {
+    delete _big_initial; _big_initial = tl._big_initial; tl._big_initial = 0;
+    }
+
+  for( int i = 0; i < tl.characters(); ++i ) shift_characterp( tl.cpv[i] );
+
+  tl.big_initial( 0 ); tl.cpv.clear();
+  }
+
+
 Character & Textline::character( int i ) const throw()
   {
   if( i < 0 || i >= characters() )
     Ocrad::internal_error( "character, index out of bounds" );
-  return data[i];
+  return *cpv[i];
+  }
+
+
+Rectangle Textline::charbox( const Character & c ) const throw()
+  {
+  return Rectangle( c.left(), top( c.hcenter() ), c.right(), bottom( c.hcenter() ) );
   }
 
 
@@ -125,7 +155,7 @@ int Textline::mean_height() const throw()
   int c = 0, sum = 0;
 
   for( int i = 0; i < characters(); ++i )
-    if( !data[i].maybe(' ') ) { ++c; sum += data[i].height(); }
+    if( !cpv[i]->maybe(' ') ) { ++c; sum += cpv[i]->height(); }
   if( c ) sum /= c;
   return sum;
   }
@@ -136,7 +166,7 @@ int Textline::mean_width() const throw()
   int c = 0, sum = 0;
 
   for( int i = 0; i < characters(); ++i )
-    if( !data[i].maybe(' ') ) { ++c; sum += data[i].width(); }
+    if( !cpv[i]->maybe(' ') ) { ++c; sum += cpv[i]->width(); }
   if( c ) sum /= c;
   return sum;
   }
@@ -148,7 +178,7 @@ int Textline::mean_gap_width( int max_gap_width ) const throw()
 
   for( int i = 0; i < gaps; ++i )
     {
-    int gap = data[i+1].left() - data[i].right() - 1;
+    int gap = cpv[i+1]->left() - cpv[i]->right() - 1;
     if( gap > 0 && ( !max_gap_width || gap <= max_gap_width ) ) sum += gap;
     }
   if( gaps > 0 ) sum /= gaps;
@@ -160,7 +190,7 @@ int Textline::mean_hcenter() const throw()
   {
   int sum = 0;
 
-  for( int i = 0; i < characters(); ++i ) sum += data[i].hcenter();
+  for( int i = 0; i < characters(); ++i ) sum += cpv[i]->hcenter();
   if( characters() ) sum /= characters();
   return sum;
   }
@@ -170,24 +200,12 @@ int Textline::mean_vcenter() const throw()
   {
   int sum = 0;
 
-  for( int i = 0; i < characters(); ++i ) sum += data[i].vcenter();
+  for( int i = 0; i < characters(); ++i ) sum += cpv[i]->vcenter();
   if( characters() ) sum /= characters();
   return sum;
   }
 
 
-void Textline::join( Textline & tl ) throw()
-  {
-  if( tl._big_initial &&
-      ( !_big_initial || tl._big_initial->h_precedes( *_big_initial ) ) )
-    big_initial( *(tl._big_initial) );
-
-  for( int i = 0; i < tl.characters(); ++i ) shift_character( tl.data[i] );
-
-  tl.delete_big_initial(); tl.data.clear();
-  }
-
-
 void Textline::print( const Control & control ) const throw()
   {
   if( _big_initial ) _big_initial->print( control );
@@ -203,12 +221,15 @@ void Textline::dprint( const Control & control, bool graph, bool recursive )
     std::fprintf( control.outfile, "mean height = %d, track segments = %d\n",
                   mean_height(), segments() );
 
-  if( _big_initial ) _big_initial->dprint( control, graph, recursive );
+  if( _big_initial )
+    {
+    Character & c = *_big_initial;
+    c.dprint( control, c, graph, recursive );
+    }
   for( int i = 0; i < characters(); ++i )
     {
-    if( graph || recursive )
-      std::fprintf( control.outfile, "track vcenter = %d\n", vcenter( character(i).hcenter() ) );
-    character( i ).dprint( control, graph, recursive );
+    Character & c = character( i );
+    c.dprint( control, charbox( c ), graph, recursive );
     }
   std::fputs( "\n", control.outfile );
   }
@@ -234,19 +255,19 @@ void Textline::recognize1( const Charset & charset ) const throw()
   {
   if( _big_initial )
     {
-    _big_initial->recognize1( charset, _big_initial->vcenter() );
-    if( _big_initial->guesses() == 1 )
+    Character & c = *_big_initial;
+    c.recognize1( charset, c );
+    if( c.guesses() == 1 )
       {
-      int code = _big_initial->guess( 0 ).code;
-      if( UCS::islower( code ) )
-        _big_initial->only_guess( UCS::toupper( code ), 0 );
+      int code = c.guess( 0 ).code;
+      if( UCS::islower( code ) ) c.only_guess( UCS::toupper( code ), 0 );
       }
-    else if( _big_initial->guesses() > 1 ) _big_initial->clear_guesses();
+    else if( c.guesses() > 1 ) c.clear_guesses();
     }
 
   for( int i = 0; i < characters(); ++i )
     {
     Character & c = character( i );
-    c.recognize1( charset, vcenter( c.hcenter() ) );
+    c.recognize1( charset, charbox( c ) );
     }
   }
