@@ -13,32 +13,35 @@
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
-    Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
 
+class Bitmap;
+
 class Textline : public Track
   {
   Character * _big_initial;
-  mutable std::vector< Character > data;
+  mutable std::vector< Character * > cpv;
 
 public:
   Textline() throw() : _big_initial( 0 ) {}
   Textline( const Textline & tl ) throw();
   Textline & operator=( const Textline & tl ) throw();
-  ~Textline() throw() { delete_big_initial(); }
+  ~Textline() throw();
   void set_track() throw();
 
   const Character * big_initial() const throw() { return _big_initial; }
-  void big_initial( const Character & c ) throw();
-  void delete_big_initial() throw();
+  void big_initial( Character * p ) throw();
+//  void delete_big_initial() throw();
 
   void delete_character( int i ) throw();
-  int  shift_character( const Character & c ) throw();
+  int  shift_characterp( Character * p ) throw();
   bool insert_space( int i, bool tab = false ) throw();
+  void join( Textline & tl ) throw();
 
   Character & character( int i ) const throw();
   Rectangle charbox( const Character & c ) const throw();
-  int characters() const throw() { return data.size(); }
+  int characters() const throw() { return cpv.size(); }
 
   int mean_height() const throw();
   int mean_width() const throw();
@@ -46,7 +49,6 @@ public:
   int mean_hcenter() const throw();
   int mean_vcenter() const throw();
 
-  void join( Textline & tl ) throw();
   void print( const Control & control ) const throw();
   void dprint( const Control & control, bool graph, bool recursive ) const throw();
   void xprint( const Control & control ) const throw();
