@@ -13,7 +13,7 @@
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
-    Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
 
 #include <cctype>
@@ -24,8 +24,8 @@
 #include "vrhomboid.h"
 #include "track.h"
 #include "ucs.h"
+#include "bitmap.h"
 #include "block.h"
-#include "blockmap.h"
 #include "character.h"
 #include "textline.h"
 
@@ -34,10 +34,9 @@
 
 namespace {
 
-int find_space( const std::vector< Character > & character_vector, int i ) throw()
+int find_space( const std::vector< Character * > & cpv, int i ) throw()
   {
-  while( i < (int)character_vector.size() && !character_vector[i].maybe(' ') )
-    ++i;
+  while( i < (int)cpv.size() && !cpv[i]->maybe(' ') ) ++i;
   return i;
   }
 
@@ -56,8 +55,10 @@ void Textline::recognize2( const Charset & charset ) throw()
     if( !c.guesses() && c.blocks() == 2 &&
         c.block( 0 ).v_overlaps( c.block( 1 ) ) )
       {
-      Character c1( c.block( 0 ) ); c1.recognize1( charset, charbox( c1 ) );
-      Character c2( c.block( 1 ) ); c2.recognize1( charset, charbox( c2 ) );
+      Character c1( new Block( c.block( 0 ) ) );
+      c1.recognize1( charset, charbox( c1 ) );
+      Character c2( new Block( c.block( 1 ) ) );
+      c2.recognize1( charset, charbox( c2 ) );
       if( ( c1.guesses() && c2.guesses() ) ||
           Ocrad::similar( c1.height(), c2.height(), 20 ) )
         {
@@ -65,7 +66,7 @@ void Textline::recognize2( const Charset & charset ) throw()
         else { c = c2; c2 = c1; }
         // discards spurious dots
         if( !c2.maybe('.') || c2.top() > c.vcenter() )
-          { shift_character( c2 ); ++i; }
+          { shift_characterp( new Character( c2 ) ); ++i; }
         }
       }
     }
@@ -78,7 +79,8 @@ void Textline::recognize2( const Charset & charset ) throw()
         c.block( 0 ).size() > 10 * c.block( 1 ).size() &&
         c.block( 1 ).top() > charbox( c ).bottom() )
       {
-      Character c1( c.block( 0 ) ); c1.recognize1( charset, charbox( c1 ) );
+      Character c1( new Block( c.block( 0 ) ) );
+      c1.recognize1( charset, charbox( c1 ) );
       if( c1.guesses() ) c = c1;
       }
     }
@@ -98,13 +100,12 @@ void Textline::recognize2( const Charset & charset ) throw()
       if( ib < 0 || 10 * c.block( ib ).width() < 9 * c.width() )
         continue;
       const Block & b = c.block( ib );
-      const Blockmap & bm = *b.blockmap();
       int colmin = 0, cmin = b.height() + 1;
       for( int col = b.hpos( 30 ); col <= b.hpos( 70 ); ++col )
         {
         int c = 0;
         for( int row = b.top(); row <= b.bottom(); ++row )
-          if( b.compare_id( bm.id( row, col ), true ) ) ++c;
+          if( b.id( row, col ) ) ++c;
         if( c < cmin ) { cmin = c; colmin = col; }
         }
       if( 4 * cmin > b.height() ||
@@ -112,29 +113,23 @@ void Textline::recognize2( const Charset & charset ) throw()
             ( colmin <= b.hpos( 40 ) || colmin >= b.hpos( 60 ) ) ) ) continue;
       if( colmin <= b.left() || colmin >= b.right() ) continue;
       Rectangle r1( b.left(), b.top(), colmin - 1, b.bottom() );
-      Block b1( r1, bm, b.id() ); b1.adjust_height();
       Rectangle r2( colmin + 1, b.top(), b.right(), b.bottom() );
-      Block b2( r2, bm, b.id() ); b2.adjust_height();
-      for( int j = 0; j < b.blocks(); ++j )
-        {
-        const Block & bj = b.block( j );
-        if( b1.includes( bj ) ) b1.shift_block( bj );
-        else if( b2.includes( bj ) ) b2.shift_block( bj );
-        }
-      Character c1( b1 );
-      Character c2( b2 );
+      Block b1( b, r1 ); b1.adjust_height(); b1.find_holes();
+      Block b2( b, r2 ); b2.adjust_height(); b2.find_holes();
+      Character c1( new Block( b1 ) );
+      Character c2( new Block( b2 ) );
       for( int j = 0; j < c.blocks(); ++j ) if( j != ib )
         {
         const Block & bj = c.block( j );
-        if( c1.includes_hcenter( bj ) ) c1.shift_block( bj );
-        else if( c2.includes_hcenter( bj ) ) c2.shift_block( bj );
+        if( c1.includes_hcenter( bj ) ) c1.shift_blockp( new Block( bj ) );
+        else if( c2.includes_hcenter( bj ) ) c2.shift_blockp( new Block( bj ) );
         }
       c1.recognize1( charset, charbox( c1 ) );
       c2.recognize1( charset, charbox( c2 ) );
       if( ( c1.guesses() && c2.guesses() ) ||
           ( ( c1.guesses() || c2.guesses() ) && c.width() > c.height() ) )
         {
-        c = c1; shift_character( c2 );
+        c = c1; shift_characterp( new Character( c2 ) );
         if( !c1.guesses() ) --i; else if( c2.guesses() ) ++i;
         }
       }
@@ -145,47 +140,105 @@ void Textline::recognize2( const Charset & charset ) throw()
   for( int i = 0; i < characters(); ++i )
     {
     Character & c = character( i );
-    if( !c.guesses() && c.blocks() == 1 && c.block( 0 ).blocks() )
+    if( !c.guesses() && c.blocks() == 1 && c.block( 0 ).holes() )
       {
       Character c1( c );
       Block & b = c1.block( 0 );
-      for( int j = b.blocks() - 1; j >= 0; --j )
-        if( 64 * b.block( j ).size() <= b.size() ||
-            16 * b.block( j ).height() <= b.height() ) b.delete_block( j );
-      if( b.blocks() < c.block( 0 ).blocks() )
+      for( int j = b.holes() - 1; j >= 0; --j )
+        if( 64 * b.hole( j ).size() <= b.size() ||
+            16 * b.hole( j ).height() <= b.height() ) b.fill_hole( j );
+      if( b.holes() < c.block( 0 ).holes() )
         {
         c1.recognize1( charset, charbox( c1 ) );
         if( c1.guesses() ) { c = c1; continue; }
         }
-      if( b.blocks() == 1 && 16 * b.block( 0 ).size() < b.size() &&
+/*      if( b.holes() == 1 && 25 * b.hole( 0 ).size() < b.size() &&
           Ocrad::similar( b.height(), b.width(), 40 ) )
         {
-        b.delete_block( 0 );
+        b.fill_hole( 0 );
         c1.recognize1( charset, charbox( c1 ) );
         if( c1.guesses() ) { c = c1; continue; }
-        }
+        }*/
       }
     }
 
   // separate merged characters recognized by recognize1
   for( int i = 0; i < characters(); )		// FIXME this block sucks
     {
-    if( !data[i].guesses() || data[i].guess( 0 ).code != 0 ) { ++i; continue; }
-    Character c( data[i] );
+    if( !cpv[i]->guesses() || cpv[i]->guess( 0 ).code != 0 ) { ++i; continue; }
+    Character c( *cpv[i] );
     delete_character( i );
     if( c.guesses() >= 3 && c.blocks() >= 1 )
       {
       int left = c.guess( 0 ).value;
       for( int g = 1; g < c.guesses(); ++g )
         {
-        const Block & b = c.block( c.blocks() - 1 );
-        Rectangle r( left, b.top(), c.guess( g ).value, b.bottom() );
-        Character c1( Block( r, *(c.blockmap()), b.id() ) );
-        for( int k = 0; k < c.blocks(); ++k )
-          if( !c.block( k ).includes( r ) && r.includes_hcenter( c.block( k ) ) )
-            c1.shift_block( c.block( k ) );
-        c1.add_guess( c.guess( g ).code, 0 ); shift_character( c1 );
-        left = r.right() + 1;
+        Block & b = c.block( c.blocks() - 1 );
+        Rectangle re( left, b.top(), c.guess( g ).value, b.bottom() );
+        b.add_rectangle( re );
+        Character c1( new Block( b, re ) );
+        for( int k = 0; k < c.blocks() - 1; ++k )
+          if( !c.block( k ).includes( re ) && re.includes_hcenter( c.block( k ) ) )
+            c1.shift_blockp( new Block( c.block( k ) ) );
+        c1.add_guess( c.guess( g ).code, 0 );
+        shift_characterp( new Character( c1 ) );
+        left = re.right() + 1;
+        }
+      }
+    }
+
+  // choose between 'B' and 'a'
+  for( int i = 0, begin = 0; i < characters(); ++i )
+    {
+    Character & c1 = character( i );
+    if( c1.maybe(' ') ) { begin = i + 1 ; continue; }
+    if( c1.guesses() )
+      {
+      int code = c1.guess(0).code;
+      if( c1.guesses() != 2 || code != 'B' || c1.guess(1).code != 'a' ) continue;
+      if( 4 * c1.height() > 5 * mean_height() ) continue;
+      for( int j = begin; j < characters(); ++j ) if( j != i )
+        {
+        Character & c2 = character( j );
+        if( c2.maybe(' ') ) break;
+        if( c2.guesses() >= 1 )
+          {
+          int code2 = c2.guess(0).code;
+          if( code2 >= 128 ) continue;
+          if( ( std::isalpha( code2 ) && 5 * c1.height() < 4 * c2.height() ) ||
+              ( std::islower( code2 ) && !UCS::islower_small_ambiguous( code2 ) &&
+                ( c1.height() <= c2.height() ||
+                  Ocrad::similar( c1.height(), c2.height(), 10 ) ) ) )
+            { c1.swap_guesses( 0, 1 ); break; }
+          }
+        }
+      }
+    }
+
+  // choose between '8' and 'a' or 'e'
+  for( int i = 0, begin = 0; i < characters(); ++i )
+    {
+    Character & c1 = character( i );
+    if( c1.maybe(' ') ) { begin = i + 1 ; continue; }
+    if( c1.guesses() == 2 && c1.guess(1).code == '8' )
+      {
+      int code = c1.guess(0).code;
+      if( ( code != 'a' && code != 'e' ) || 5 * c1.height() < 4 * mean_height() )
+        continue;
+      for( int j = begin; j < characters(); ++j ) if( j != i )
+        {
+        Character & c2 = character( j );
+        if( c2.maybe(' ') ) break;
+        if( c2.guesses() >= 1 )
+          {
+          int code2 = c2.guess(0).code;
+          if( code2 >= 128 ) continue;
+          if( ( std::isalpha( code2 ) && 4 * c1.height() > 5 * c2.height() ) ||
+              ( std::isupper( code2 ) &&
+                ( c1.height() >= c2.height() ||
+                  Ocrad::similar( c1.height(), c2.height(), 10 ) ) ) )
+            { c1.swap_guesses( 0, 1 ); break; }
+          }
         }
       }
     }
@@ -308,24 +361,31 @@ void Textline::recognize2( const Charset & charset ) throw()
     }
 
   // transform a small 'p' to a capital 'P'
-  for( int i = characters() - 1; i > 0; --i )
+  for( int i = characters() - 1; i >= 0; --i )
     {
-    Character & c1 = character( i - 1 );
+    Character & c1 = character( i );
     if( c1.guesses() == 1 && c1.guess( 0 ).code == 'p' )
       {
-      Character & c2 = character( i );
-      if( !c2.guesses() ) continue;
-      int code = c2.guess( 0 ).code;
-      if( code >= 128 || ( !std::isalnum( code ) && code != '.' && code != '|' ) )
-        continue;
-      switch( code )
+      bool cap = false, valid_c2 = false;
+      if( i < characters() - 1 && character(i+1).guesses() )
         {
-        case 'g': case 'j': case 'p': case 'q': case 'y':
-                  if( c1.bottom() + 2 > c2.bottom() ) continue; break;
-        case 'Q': if( std::abs( c1.top() - c2.top() ) > 2 ) continue; break;
-        default : if( std::abs( c1.bottom() - c2.bottom() ) > 2 ) continue;
+        Character & c2 = character( i + 1 );
+        int code = c2.guess( 0 ).code;
+        if( code < 128 && ( std::isalnum( code ) || code == '.' || code == '|' ) )
+          {
+          valid_c2 = true;
+          switch( code )
+            {
+            case 'g': case 'j': case 'p': case 'q': case 'y':
+                      cap = ( c1.bottom() + 2 <= c2.bottom() ); break;
+            case 'Q': cap = ( std::abs( c1.top() - c2.top() ) <= 2 ); break;
+            default : cap = ( std::abs( c1.bottom() - c2.bottom() ) <= 2 );
+            }
+          }
         }
-      c1.only_guess( 'P', 0 );
+      if( !valid_c2 && i > 0 && !character(i-1).maybe(' ') )
+        cap = ( std::abs( c1.bottom() - charbox(c1).bottom() ) <= 2 );
+      if( cap ) c1.only_guess( 'P', 0 );
       }
     }
 
@@ -384,7 +444,7 @@ void Textline::recognize2( const Charset & charset ) throw()
   // transform words like 'lO.OOO' into numbers like '10.000'
   for( int begin = 0, end = 0; begin < characters(); begin = end + 1 )
     {
-    end = find_space( data, begin );
+    end = find_space( cpv, begin );
     if( end - begin < 2 ) continue;
     Character & c1 = character( begin );
     if( !c1.guesses() ) continue;
@@ -419,7 +479,7 @@ void Textline::recognize2( const Charset & charset ) throw()
       }
     }
 
-  // transform a vertical bar into l or I (or a l into an I)
+  // transform a vertical bar into 'l' or 'I' (or a 'l' into an 'I')
   for( int i = 0; i < characters(); ++i )
     {
     Character & c = character( i );
@@ -436,12 +496,12 @@ void Textline::recognize2( const Charset & charset ) throw()
       if( i < characters() - 1 && character( i + 1 ).guesses() )
         rcode = character( i + 1 ).guess( 0 ).code;
       if( UCS::isupper( rcode ) &&
-          ( !lcode || UCS::isupper( lcode ) || UCS::isspace( lcode ) ) )
+          ( !lcode || UCS::isupper( lcode ) || !UCS::isalnum( lcode ) ) )
         { c.insert_guess( 0, 'I', 1 ); continue; }
       if( code == 'l' ) continue;
       if( UCS::isalpha( lcode ) || UCS::isalpha( rcode ) )
         { c.insert_guess( 0, 'l', 1 ); continue; }
-      if( rcode == '|' && ( !lcode || UCS::isspace( lcode ) ) )
+      if( rcode == '|' && ( !lcode || !UCS::isalnum( lcode ) ) )
         {
         if( i < characters() - 2 && character( i + 2 ).guesses() &&
             std::isalpha( character( i + 2 ).guess( 0 ).code ) )
@@ -533,9 +593,18 @@ void Textline::recognize2( const Charset & charset ) throw()
     if( c.guesses() )
       {
       int code = c.guess( 0 ).code;
-      if( i == begin && code == 'a' && c.guesses() == 2 &&
-          c.guess( 1 ).code == 'Q' && 4 * c.height() > 5 * mean_height() )
-        c.swap_guesses( 0, 1 );
+      if( i == begin && code == 'a' && c.guesses() == 2 && c.guess( 1 ).code == 'Q' )
+        {
+        if( 4 * c.height() > 5 * mean_height() )
+          c.swap_guesses( 0, 1 );
+        else if( i < characters() - 1 && character( i + 1 ).guesses() )
+          {
+          int code1 = character( i + 1 ).guess( 0 ).code;
+          if( ( std::isdigit( code1 ) || UCS::isupper( code1 ) || code1 == 'i' ) &&
+              5 * c.height() > 4 * character( i + 1 ).height() )
+          c.swap_guesses( 0, 1 );
+          }
+        }
       }
     }
 
@@ -551,34 +620,6 @@ void Textline::recognize2( const Charset & charset ) throw()
       }
     }
 
-  // choose between 'B' and 'a'
-  for( int i = 0, begin = 0; i < characters(); ++i )
-    {
-    Character & c1 = character( i );
-    if( c1.maybe(' ') ) { begin = i + 1 ; continue; }
-    if( c1.guesses() )
-      {
-      int code = c1.guess(0).code;
-      if( c1.guesses() != 2 || code != 'B' || c1.guess(1).code != 'a' ) continue;
-      if( 4 * c1.height() > 5 * mean_height() ) continue;
-      for( int j = begin; j < characters(); ++j ) if( j != i )
-        {
-        Character & c2 = character( j );
-        if( c2.maybe(' ') ) break;
-        if( c2.guesses() >= 1 )
-          {
-          int code2 = c2.guess(0).code;
-          if( code2 >= 128 ) continue;
-          if( ( std::isalpha( code2 ) && 5 * c1.height() < 4 * c2.height() ) ||
-              ( std::islower( code2 ) &&
-                ( c1.height() <= c2.height() ||
-                  Ocrad::similar( c1.height(), c2.height(), 10 ) ) ) )
-            { c1.swap_guesses( 0, 1 ); break; }
-          }
-        }
-      }
-    }
-
   // join a 'n' followed by a 'I' into a 'm'
   for( int i = 0; i < characters() - 1; ++i )
     {
@@ -595,4 +636,21 @@ void Textline::recognize2( const Charset & charset ) throw()
       }
     }
 
+  // join the secuence '', '/', 'o', ' ' into a '%'
+  for( int i = 0; i < characters() - 3; ++i )
+    {
+    Character & c1 = character( i );
+    if( c1.guesses() == 1 && c1.guess( 0 ).code == UCS::DEG )
+      {
+      if( character( i + 1 ).maybe('/') &&
+          character( i + 2 ).maybe('o') &&
+          character( i + 3 ).maybe(' ') )
+        {
+        c1.join( character( i + 1 ) ); c1.join( character( i + 2 ) );
+        delete_character( i + 2 ); delete_character( i + 1 );
+        c1.only_guess( '%', 0 );
+        }
+      }
+    }
+
   }
