@@ -1,5 +1,5 @@
 /*  GNU Ocrad - Optical Character Recognition program
-    Copyright (C) 2003, 2004 Antonio Diaz Diaz.
+    Copyright (C) 2003, 2004, 2005 Antonio Diaz Diaz.
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -13,7 +13,7 @@
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
-    Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
 
 #include <algorithm>
@@ -84,7 +84,7 @@ int Vrhomboid::vcenter( int col ) const throw()
   {
   int dx = _right - _left, dy = _rvcenter - _lvcenter;
   int vc = _lvcenter;
-  if( dx && dy ) vc += ( dy * ( col - _left ) / dx );
+  if( dx && dy ) vc += ( dy * ( col - _left ) ) / dx;
   return vc;
   }
 
